"""ledgercomm.interfaces.hid_device module."""

from typing import List, Tuple, Optional

try:
    import hid
except ImportError:
    hid = None

from .comm import Comm
from ..log import LOG


class HID(Comm):
    """HID class.

    Mainly used to communicate with Nano S/X through USB.

    Parameters
    ----------
    vendor_id: int
        Vendor ID of the device. Default to Ledger Vendor ID 0x2C97.

    Attributes
    ----------
    device : hid.device
        HID device connection.
    path : Optional[bytes]
        Path of the HID device.
    __opened : bool
        Whether the connection to the HID device is opened or not.

    """

    def __init__(self, vendor_id: int = 0x2C97, hid_path: Optional[bytes] = None) -> None:
        """Init constructor of HID."""
        if hid is None:
            raise ImportError("hidapi is not installed, try: "
                              "'pip install ledgercomm[hid]'")

        self.device = hid.device()
        self.path: Optional[bytes] = hid_path
        self.__opened: bool = False
        self.vendor_id: int = vendor_id

    def open(self) -> None:
        """Open connection to the HID device.

        Returns
        -------
        None

        """
        if not self.__opened:
            if self.path is None:
                self.path = HID.enumerate_devices(self.vendor_id)[0]
            self.device.open_path(self.path)
            self.device.set_nonblocking(True)
            self.__opened = True

    @staticmethod
    def enumerate_devices(vendor_id: int = 0x2C97) -> List[bytes]:
        """Enumerate HID devices to find Nano S/X.

        Parameters
        ----------
        vendor_id: int
            Vendor ID of the device. Default to Ledger Vendor ID 0x2C97.

        Returns
        -------
        List[bytes]
            List of paths to HID devices which should be Nano S or Nano X.

        """
        devices: List[bytes] = []

        for hid_device in hid.enumerate(vendor_id, 0):
            if (hid_device.get("interface_number") == 0 or
                    # MacOS specific
                    hid_device.get("usage_page") == 0xffa0):
                devices.append(hid_device["path"])

        assert len(devices) != 0, (
            f"Can't find Ledger device with vendor_id {hex(vendor_id)}")

        return devices

    def send(self, data: bytes) -> int:
        """Send `data` through HID device `self.device`.

        Parameters
        ----------
        data : bytes
            Bytes of data to send.

        Returns
        -------
        int
            Total length of data sent to the device.

        """
        if not data:
            raise Exception("Can't send empty data!")

        LOG.debug("=> %s", data.hex())

        data = int.to_bytes(len(data), 2, byteorder="big") + data
        offset: int = 0
        seq_idx: int = 0
        length: int = 0

        while offset < len(data):
            # Header: channel (0x0101), tag (0x05), sequence index
            header: bytes = b"\x01\x01\x05" + seq_idx.to_bytes(2, byteorder="big")
            data_chunk: bytes = (header +
                                 data[offset:offset + 64 - len(header)])

            self.device.write(b"\x00" + data_chunk)
            length += len(data_chunk) + 1
            offset += 64 - len(header)
            seq_idx += 1

        return length

    def recv(self) -> Tuple[int, bytes]:
        """Receive data through HID device `self.device`.

        Blocking IO.

        Returns
        -------
        Tuple[int, bytes]
            A pair (sw, rdata) containing the status word and response data.

        """
        seq_idx: int = 0
        self.device.set_nonblocking(False)
        data_chunk: bytes = bytes(self.device.read(64 + 1))
        self.device.set_nonblocking(True)

        assert data_chunk[:2] == b"\x01\x01"
        assert data_chunk[2] == 5
        assert data_chunk[3:5] == seq_idx.to_bytes(2, byteorder="big")

        data_len: int = int.from_bytes(data_chunk[5:7], byteorder="big")
        data: bytes = data_chunk[7:]

        while len(data) < data_len:
            read_bytes = bytes(self.device.read(64 + 1, timeout_ms=1000))
            data += read_bytes[5:]

        sw: int = int.from_bytes(data[data_len - 2:data_len], byteorder="big")
        rdata: bytes = data[:data_len - 2]

        LOG.debug("<= %s %s", rdata.hex(), hex(sw)[2:])

        return sw, rdata

    def exchange(self, data: bytes) -> Tuple[int, bytes]:
        """Exchange (send + receive) with `self.device`.

        Parameters
        ----------
        data : bytes
            Bytes with `data` to send.

        Returns
        -------
        Tuple[int, bytes]
            A pair (sw, rdata) containing the status word and reponse data.

        """
        self.send(data)

        return self.recv()  # blocking IO

    def close(self) -> None:
        """Close connection to HID device `self.device`.

        Returns
        -------
        None

        """
        if self.__opened:
            self.device.close()
            self.__opened = False



# Copyright 2019 Shift Cryptosecurity AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""BitBox02"""

from abc import ABC, abstractmethod
import os
import enum
import sys
import base64
import hashlib
import time
from typing import Callable, Optional, Dict, Tuple, Union, Sequence

import ecdsa
from noise.connection import NoiseConnection, Keypair
import semver

from .devices import parse_device_version, DeviceInfo

from .communication import TransportLayer
from .devices import BITBOX02MULTI, BITBOX02BTC, BITBOX02PLUS_MULTI, BITBOX02PLUS_BTC

try:
    from .generated import hww_pb2 as hww
    from .generated import system_pb2 as system
except ModuleNotFoundError:
    print("Run `make py` to generate the protobuf messages")
    sys.exit()


HWW_CMD = 0x80 + 0x40 + 0x01


class HwwRequestCode:
    # New request.
    REQ_NEW = b"\x00"
    # Poll an outstanding request for completion.
    REQ_RETRY = b"\x01"
    # Cancel any outstanding request.
    REQ_CANCEL = b"\x02"
    # INFO api call (used to be OP_INFO api call), graduated to the toplevel framing so it works
    # the same way for all firmware versions.
    REQ_INFO = b"i"


class HwwResponseCode:
    # Request finished, payload is valid.
    RSP_ACK = b"\x00"
    # Request is outstanding, retry later.
    RSP_NOT_READY = b"\x01"
    # Device is busy, request was dropped.
    RSP_BUSY = b"\x02"
    # Bad request.
    RSP_NACK = b"\x03"


ERR_GENERIC = 103
ERR_DUPLICATE_ENTRY = 107
ERR_USER_ABORT = 104

HARDENED = 0x80000000


# uncompressed secp256k1 pubkey serialization, hex-encoded.
ATTESTATION_PUBKEYS: Sequence[str] = [
    "04074ff1273b36c24e80fe3d59e0e897a81732d3f8e9cd07e17e9fc06319cd16b25cf74255674477b3ac9cbac2d12f0dc27a662681fcbc12955b0bccdcbbdcfd01",
    "044c53a84f41fa7301b378bb3c260fc9b2ff1cbea7a78181279a8566797a736f12cea25fa2b1c27a844392fe9b37547dc6fbd00a2676b816e7d2d3562be2a0cbbd",
    "04e9c8dc929796aac65af5084eb54dc1ee482d5e0b5c58e2c93f243c5b70b21523324bdb78d7395317da165ef1138826c3ca3c91ca95e6f490c340cf5508a4a3ec",
    "04c2fb05889b9dff5a9fb22a59ee1d16bfc2863f0400ddcb69566e2abe8a15fa0ba1240254ca45aa310d170e724e1310ce5f611cada76c12e3c24a926a390ca4be",
    "04c4e82d6d1b91e7853eba96a871ad31fc62620b826b0b8acf815c03de31b792a98e05bb34d3b9e0df1040eac485f03ff8bbbf7a857ef1cf2a49a60ac084efb88f",
    "040526f5b8348a8d55e7b1cac043ce98c55bbdb3311b4d1bb2d654281edf8aeb21f018fb027a6b08e4ddc62c919e648690722d00c6f54c668c9bd8224a1d82423a",
    "0422491e19766bd96a56e3f2f3926a6c57b89209ff47bd10e523b223ff65ab9af11c0a5f62c187514f2117ce772de90f9901ee122af78e69bbc4d29eec811be8ec",
    "049f1b7180014b6de60d41f16a3c0a37b20146585e4884960249d30f3cd68c74d04420d0cedef5719d6b1529b085ecd534fa6c1690be5eb1b3331bc57b5db224dc",
    "04adaa011a4ced11310728abb64f09636267ce0b05782da6d3eeaf987cec7c64f279ad55327184f9e5b4a1e53089b31bcc65032dad7205325f41ed3d9fdfba1f88",
    "044a70e663d7fe5fe0d4cbbb752883e35222b8d7d7bffdaa8d591995d1252528a4e9a3e4d5220d485021728b3cdad4fccc681a6ddeea8e2f7c55b4acde8d53573d",
    "04ef42d5c74254dd6afb36ec752068252ed6b6e231d019b0dfe32ee9ac2d5444471698e0ce7626e2f1f6266f42e343a20fd6fe2fd9f206e623c2eb6c1c9922465d",
    "04438a5d3a6262e631ad7a59ae9ffa0ee73b68957c4b3ad46d5747cae81beb15beafd1feed9c50fbff7f005ae649181c987649e1b251db74a052fef0b6e99a8064",
    "04d145a6ba3bf76d5db1b80dbeaef8c66a4eea215d1b6f7260a0b2b63fd30c3f4c3e92913a6ce272b36ff622fff64a67b5cf8ef5db645f403c2647d4c11d5f968e",
    "04458f68edde1d059ef2ccc970938cd44f012da4a8556f35f5a4e1df87313c7adfcd4f22e77345e5cc900eb6dd48af9bcb42bebe3163a79b3a3a3d32c359b5c06d",
    "0456bb82ddc84bd87d0d6dca27439dad6f0901518bdea2347c1c07e1a53a8480ec24b615af1d6cdb9344d8ea8b2103c9daa674742c260b1d2532acc07cb1b4fce6",
    "04f9d9ab547dce5b4d1c4778dae1990746e9e1f9a02475b46f88d8d4cc901db348988be3a5b8466a89f5a55e75416912d4b109eed86298577a718c0e3a023b0bf2",
    "040682f7a78d6290e6c59e9d1bde9cd665fa24232198385a10948906b473c59b160f95308ac8adf092276722d68f9db77d61df72587797ee8c3cf0df45ae19a0ba",
    "0432b2b5f54365847dcd1a9c96abb041a3d6483e93049fc3440b7b46ef8bac8d397d042dd48bb9928de3f2e343a670792fe25f65e5ffcbbf0121c4a0a061e55cd3",
    "043eef9f2af390dfb456800fc7673f1ac61baefa68bff84ccd36fdfe408f90205b865ddd2c6036ceb6260bd6e6b2f2b0a0ade9a60579f259bcb945ea5eced8bfe8",
    "04bc9b0fee1e5e088c77608e6ce16313a96a5b725de2c77fb3504e8c6ecf54ba278837f6ecdeb5a537d89d0a655cdaface33ef7786b949533ea8cec90694b15840",
    "046f8f4942d0ffea7078850917b850b96fac0d1010ff9ef1419e5e9a8fce94796ff1c451f64ec8e5fbb20bc1cb73fd5fdddf5c647883f3966a82ed3e816d8da862",
    "04995f9c5d5f539e53cd106837ca3eba50626df5dd5e54ce3146477c57cd5918c0b83b30f3e469c8ee7c741cd7c6833e78eeab710dfb96360ce464905fbda8179e",
    "049991c5db649d7cf37f0a1cae2b4f14461404e0286c0b86e3dd5462a18a3ec35f54e66865a8e88aab14f94cbd8778c3c79440c54e5eeb5f96ba097638017b71dc",
    "0433b669b2346c99824e81352a254ef4688349f41570f531e031a9c51e901aa0d74dc1206e0bd41929a75e44ee3e842c9d746f93172f5d9e3dd7a22e79d0fbad1c",
    "04d2b99da82c7da93db72b6e6ab52dad09bd71d661e6ff63f22170afb4599fd88456aa8f555d910c5d4cc5a8cffb38f0cd7ec514aeecf3af990dc3ea7ba9d0d1ff",
    "043be5abaf680e97f416a66538f0196f9ac8a2efc4f5ddc73ba43b6b3c261216b10f08b4c8c37a63dea0d2fa05240408f6d3de6d9f730af2c4e890fe1256d134aa",
    "04f9b88dcff16de9f4d0372e36dbf220cec95099041003ed4f3a894bc6ffd0e881ca45f3c5dd0518d14d98e7e377623f5c4d99f302ce482b0a1ca0c6f046f5d251",
    "04564c4238efbaca156d159ec945b80b7979dce14c4a59d380b1fc242d3aa3de0ad4ddae7262a11169819913e4346cb89b1ddc427d2192abd8f6c5a2c850768ba1",
    "0486fb44812e9284ac4843463021729f33da45f0b90e3947d642f84320fd1be68deb24989b48b0a8d5c2ce438693ec1405d2358198fa939ea1ecd52b0df1fbc498",
    "04a4b9b7a03abd093543b2d8dddddead13be13e7529ce0202b5b29a8ac52d0cfabb97bae52301a8a0e4853f3220e69dd2a0abfce526fdc087e7e0463e6495f739f",
    "04fb09aa6b917f136a221d299e4d021adbb309d4753f94fb502f4b6e87abeb475221180c3467762f8342a1edb55a871c016262c1a8526a025a9bb3035efbf1c566",
    "04f7b1dfff6f5b91432c6fdc0f79d13017b95d03d12e8b688ac446f89f9e5749d056ea2ea32d4f43cf12b174e2f7dbbd35c86d99891fd15512dddc7813ce710d6a",
    "04476ed4537511dd1f24588bfd74e3e34d0d4158d7042037abe6ad86e91beac3b29a4db33c3eeffeaf8e41c86996da2992a0a168bd7f8f6298a6f0c592046635dd",
    "04e6a3ba1a4024d7f7a666af76c880f3285bad6e945416b06ba66c6ad2c9ba0344315136d4d2a398c9142ec205f844b896e2eeb2060a3ad1108879f811040902b7",
    "04f6d29602c85fbc3af84f295face5b10e996520d19e8e3b95e71a56c30d6c868c38e11df16293c304ccb9cb6e0d4c7d9e7f5badd488e6747eaf88d0be221319fe",
    "04517ad3854bf3b9fa52ff703bd6d0fefc61406a7cd606cff270fc8ea1233366e5592321a8bb4dbbaca4df3a80ad68ef9de92e245aae898616d0ababad5f2449a6",
    "04e1d94a45345ff33512c63ed5a6bb2f3fb92cc1db15267389cad9e8ef70de3db522faad29ad06f2a73f174b12a23e49c272e673f28a710243bfa8adc2ffd12f1a",
    "045570fa35433648d7a6e98274629c3dc565e3211331802baa64087d12a63c53488071a619dd0313993d423a65a2175baba1abdcd5df7186ed247a9eaceb8b74fd",
    "0401fe456142c8fa3b15f7d381fc62b40dcc546d5786aaf5fd08d8b2963c90895096ce773dff54b41ef604192ba9dd48786319a4c53e22217200c20cf3fcd95d6b",
    "04c552738db26ef0e89bed1be8ade4951e7734579f5c49954373e78ae5e613235835e7835255eb92b64dba9d52db768895d8890e1f60f95522127cbfdf501581a8",
    "040366885c11552f8354a6d6d133c1d3a1d4486ac5b85d94733f583d494d39aa05f54bb0d47746bcc082ebdf7f317f33df32e517982a3e1454fbc4b776c012ce79",
    "04fb246edb986718ac4e892c336fd3100cd225f389806042d9fdbff1ac84be110ee1b9cfa38c4bd7c7334c8b8d21b7d07e00a1b9910b6b103b8b83cf4f27b352b4",
    "046003d369e24ce07356355c3bc0326856412b9d2259686e0f391b223305a7a3894b2cfe9285752865326d47fce223ea24d3c6a087fefd2a80f38041d6406da5c2",
    "044481ac1f9fd76dc8d87d4df88944fcc974c8ef4a4283610454eaf67393c38bcc21459132d90d115d208afbdf4bf77d3dede3e904804339a7fce5e57f0ab1bc0b",
    "04c7b2d16a898ba1233ff19a3bb1ce8c8956f2db8b42b0ed57bc1b1d249d1e2cb53a800a18050e44bb2d82d48066fc617dd81a0739dc71957a3daf6f22e30d5b39",
    "046d3e716eb3580f5335d9d3869ea048c1b4c065817716f2edd2dcb8cefbe9f6d53fe3afd76d43ee48c4b4467fee60685a0afe555ce2b65a2a6a96768f05601bc4",
    "044473a7c536402aa209b666acaa95f28379916cbaa6a2f4f03f544698e590e917678b102bfb08a8bb31a8d7ef4f82508373a42903d5bf4b267e4e0d285da81d75",
    "0451e34fa15a57dfa8c1bed92d965a8e9ad58531f79851f7aca792f3ae55e83daed5ff4b9552bd9b49eefb6f1aca173cde9b9469e7df3194a089b8f4d833305398",
    "0434aac7b381d143f4588e5018ec956d8dfcbca4b9e04d17e0c9abd96c5a88b562cc816c8ca24a0c50550df80e992801e2fc3427ae15ff5857083319686019d4b5",
    "047560b4430a1fa56ff42e47fd5f4063f39e993a7af0f04d13a75d601ca42508ba11c3245362763fb0c7847002d364a4d7913052131c11c1c7056330c12e9c29aa",
    "0408a651514bf9e992d297629be1f4c1ef7804760451bad13daf701fc3c662f2297fdd166cb8440c6d78527926d71664b1eb5a88ac5b75fd7a9aebfa30a2ac2db1",
    "04b6be971ec9f51de953e9689a9d7df828f45459f78cbfbc6f9bdab8a1d5f7bf6305991207c2f3fc48695f113936eac5599130b3bd4bf3f8ed125888c106bf6675",
    "04a7436fd434513ec3748e01dba2f1cc6406a945a0329cfd3e60fa04fc8196d21c83a8e34d688afb1a29f251ef29ea659849eac310c8bc38ec29edc23b40cd860e",
    "041e432d24bed03c5cfddb38155e2df8e42482294423c79a0341f0b86073d91cea71ffc58bacaf08c11511461e000ecd6db33170ee99e1aca8aff89f68b7c5db31",
    "045604b119b5d7ce91093a429c70a246842d9150390eaf5e98530fdd68ff0263051159d4baed1eacde5cce3064a061910348ae3e5dc88b9daef5417418b7fdb57d",
    "04aa59b2506202e2112e9bb794cfe43aace1c303559332869e04a151a2b2dc70d382b9b156cfd5a710ff476b5e4744439e7763e076f7dd834ef21900d949798f93",
    "0433f5b6c27f93489a98c3a109cae3da0ea86b93d75cf8871f3f78e9696b4045a7342ae38ac06359f710aa2c06a88821ac79672c2498fbec051e4f1525e8c9fae7",
    "047cbbe361c84de6b9c9c809b12a584354780b12aa89787673cfc2ca22d48bb0eb43a8aac781d333b8b91956be82840985cc90fd9ba8443589661b6aa308478ec6",
    "04768cdf5973c34f027abb68e86844eb981b3ddb5a5f11f8c079b04fe658afeccc8163cab7e97120ada6313a2c3007a948a1ea1b3d466106cf472a127974a971ce",
    "041bc1b519a2a67f14c91bc8052cc3d94a7a2824d742abb1be65c7d86085ed0b15c678ff508b01add8bc5f24295f4bfbfb0126c3bbb05ba0bb596062ba37f86c43",
    "04ec87da7c538369614ead4d3a3b359fa22468fac84a101623f2394eea10c8ab92b2f118951333a35311e11cffa2ecb9f7a9b9d5e1360efc636fc8b1be830c812b",
    "0487cb0069f2a43cd3c0be0d5e6164bb2a4d56db2bd962b25b7159bff04df286010b4e2c54982ee72f573093dede1ef7e7b337bd0a0181f270098cc3aeeb743b5a",
    "0487e18223782642f5605fbc2cab0cb3d78a35367da75bc3466aef5fce03375fb36bb24fe3653da9085fde2db75624a2283866f0d39b788ad9538a7795779cffdb",
    "04d3b453084456096bc30dd5ac28d6cbab06ca40424543fcc72f3becc198e4cb14728bbd52cc452eb7417729c0e5283277c3a17a5d2faa4e26d40880560734d19b",
    "048bf5abf82e6fd10f3516f0586c0cf091941fd17411fda4333bbe8675c97f6395b5cf4e0b6c32c21005a75e3c4c950abd0c34fb8c8722d47cf17b3fb71fafb3dd",
    "04e123e5639ae797edb389875534b223439c8b9400b32d799df3f2e3319c87a2bd527f0fbe7bed07186f78b591503abf453af5a0016984176764396ff47466616d",
    "049767f3b31fd667975831ff39964e38eda53a1a9bad3ae0807d29d3929ab379cfd27d4cb935f9c089829d6a51c783ca8e73b06cec195437ff58d80ed11879291b",
    "04e95dfd536e3ed24c55a76b55cbaf74fe1aceb8b632facfad1205d608cbd0282d77d45868ff49d31afc017ddb17ddb168535929faa0d23203da71546b9cbffef6",
    "04853455c8cf75da9e19c8ee89e73057efc9d4cd8008da4c08e2a5903deb712bef3c9d2bb0a853a3a2e2d7b10a8730254babd89639c237dc54ed649e4888fb5a8e",
    "04ee55299800e7739b5673029a1fb8d88976e81442c71e2b00ec19665c6578d33d960edd546c263614f4a4f1ddfee6c916022b85f5fe5a3f675d49b65d2899ed21",
    "04813b0d011e25ccea0a6c277d8032eaf7c4f055d89e8c4b53c1ada83554f41f0c864681c6b36112a9f47c6512e5033fbfac9c5caaad2bbca15baacbb092c07b6b",
    "04a800befd176bc49d7bba6d2ec9807ee72dea039377f7b027a4aab09ecc12a0fafe80394d346a8bf1a3a4a2c7742f9fdb8814a3d01d36d9e1dc90340fb6aa7a4e",
    "04fb15f204f7c47fa21bde4ff365938071ea402816048819bddedfca8264d020f0646ac5c0cf47df85794b4dd0a17b788bdd09d45582e8e6b9562a538769f2475d",
    "04bb7f11a51e91025681d9d329dbb1fd4ee4ea1eba6afb9c89022dd555ce442331bed3215001afe6d7bcdff5a8215a4e03f3b790a7e47182184a64ad13d0d425bf",
    "0466a3f9e7a7eb51ac22e46c972bcefc4ae1b3cfcbc4a532b4d7e51ec4bbb002ce8ed3af4a94fd6fb4fcfafa2252caf328d6a7ddf16edf76079619cbe4a24d89d5",
    "04b9f07bb4ede8aed008d65ba6baa5c8743239cc68bb7b3471e290eaaff71601453932f4f0321cb761cd8f2694ab16e405297c4ad9ede751310a09978de7b008be",
    "04949f487a41fa7d9b0368a89eb9e16c885b937b08412fe6565588837862537af37a2bc7c7009af2e921273a3133e858147509b415b2f49e9f8da1f69a49175814",
    "043548e330caaf10c4696f2a712fb5b2b69cd6316de64d4405c5f2f3d74bfa44b52a72a68a58856677d2b0fb20541766236432f0807d6c2e7e897a2d508fab8b80",
    "043aa576a8171b73b0f653532c0b0f6f49b893045b80f99e2214f207093b263f173d467c9c542e4a9cc2b0dd1f6dc46238084c329612b6a54c325fa00bbaadfc59",
    "04c982547397aab0b991bb0200336c810a7b7247fca76615012b1be8307d1763617409fcedb531d44ec83dfc25310f38a46ddb83112768dcb8e31ce34efde31d17",
    "04e45e1b9dfc8ca9a03c0126b6e07fadb38b2029b03f97ee2e3c9e7f2475aa961cbf6e8c9afbe6086b321bc225a3e44b3157185c6ebe21130cad1b346712ea1716",
    "04cd33b47eff77442eb5ab5f67869efc470c62097970b66d0d74fe7ca7fd70eb76b0adc7dbe6788ac5f42fe753750c9743e7f4d0d488963a01840c3a66aa038317",
    "043370ff050a99e431726ffee05cbc95225d096ecca1e534516e7c40e741943f6410eff44af418743a117d9b174451685a4eaad2a603645d49181369d54f81714a",
    "04cf6bd3c4bbeadfa939957aa3e0ed7ace55a030a85698a8876976a69f65e5843d9332200a242957473d3871a79e90d3e3e902cb749f999e9ba3533704bbf98c8d",
    "048b8ea186a6963aa7460cb144622daf28deecf132e38d5ec26d82464c27ebf93e16915eb953bdd1d29487be55be1a566e1a904955f5fc700e03e6bb2c55e9d30d",
    "04fb36dfe7949cbcad863b2f3c7c71b1bda01c954e8211103eca93153d44be5a4bc329df2543da1767263a2e8b6596a427a55ba735aff33249dcfec89d8fc8fb49",
    "04e8b5c26cae0c7719df23b88a80ab67cd418f929a932b4ee49a52049cce0968838c805942005006bb49431676a08b653440c857a6135b098c7f30d62162240d9d",
    "040867f9cd46ec9d0ff098303dca4d3297feee1d00b20da73d9cf2e6ba0a9e83f596a6ed882401cbfedb345c105a7c5f804d683a7714d0bae4aa7e027377fb0ca8",
    "04ceedd2c63a20ca20d9f84f83f1d27617722ecba021ef217fad0d1ad2f58b76006a32e8c7ea53f7657a5e1f6f4ea1a4a55081a3670bc7ed8fa2bde2cff465f12a",
    "049947a8d9573beb26f61f499da8435fbaa69e020145a5f1fc2efa1558db4747056cd36e299f92fbb9c98f3c7cbf04985fb7c09fac83be3bac2e4b3078129e5383",
    "04082483f5c4d03f74365fd66824bd02f7be1461c3559265de71f7702f16faccee4de0bc2e5ea8dc05461535b8ec476b31b1a447744b4e6d86b98f7a5eced629b9",
    "042b892b394e21f31de43e6aaa899d158458da5778287e50815358eb25dee9cd598e8d2071519437a9d0c253a2b6fcfbd6d9df32ed7c1df9df431c552e86caf125",
    "04ef30a5877fc417b7aa8ea21f1bf76160028a49a0cd43e524a3d9bc44496233b1c5773c0b58fec30d9652ca3ca538b20310749949d38e2d1644cd7e3809b33cc1",
    "04676f9723819f36518ad34e7b21d69f830b43fb966c0a3e8d1116bd3fa061fea5099882afbed5bf343ae4b95b4b0a13622c229ea0a555d2c478686096f568a647",
    "04abb4363bdfc5a91178ddc8a4cdc05d2e0616783594c53d5af2c2d34caa9b787e685434097fe4fff5af149acfaa9ab57a8060ef9d260c0bfba0eabdf9fad40c4b",
    "043220fddfbf5ae6bc63ff750f5f040a2a6e5cd8817a7450ae48653b6a1489fbaf86c9f42d6e31c4be25e2ed9fc1ca68114d01f001a36bde776498e70377d7327f",
    "0417a20da4ae93c3532afdef5f5a29124080eb98091214c9520dc60dd2c59d14a45561bc39247cba51edad190f34551309948100d5032afab9862b7ea0c60959c8",
    "04f378380a8aee1472449e94452ef366305abee4f797298e80398c0becc339ee9a1a3a4df34af6f10703d3816b377f0e3577ec977dc2158ab25e6aec18d1b92e45",
    "049a2763f23933d8cb5664f46a4b1c62cb55087d21a655a8cb271104a3b973c5e4b4c495a3bd834b672e98206141be46f3e823ba4e20b18b33de8352c544a9b5b9",
    "045f6f1d56a429497360a87958efb92a0a17c1921ce5b85db163aae22fc428129da8bcace8812a0e0b3f80935f4ea0c71f53c089df024906f09630cc4dcbae69d2",
    "04b397d8873b66eb2778281c0b1782e9c173de6021fe7d446de6d0aaddc3946fcd6f711c5e84d3548c28cd11bab548504ac1dd25c7960a0cfb1a09c45ecf88593b",
    "04c91160f925c710c3ba4befe7e6ab22692a633d5865e02dd49ad543ecc03d78a180b32bd1c10a1e37c0c2d8439eee0a9065113a3dc9639807306d360d3c86c639",
    "04a94517c095528cf6a7b65adbd1f6723b2840392a587630b8f1089b4a744fbf026f3f594c1b0ea7d1f4d70c7648ebde515aef4eeacacbcb46287e086baaabec42",
    "049df689158d44e061b5362fe40549fa6b2231e64d01c3fe43ccb8566bbf40178f501a9aa6a3c45dfbcb6a9212a19e68c12bf18788cab71b725d44129c5ae65d62",
    "040fead3b041184e24b1bb461c6accf1b35cf63c55bc68e2eea77060331bb14c680c8065cd8a4ee1d8af2f8f8b295c982570c9aa4995f00bf26aac3b04975aab83",
    "04fce43a145bfdb469d1e45134ff68254fe260efc9b58c3c6b367a9ec587b379c0547197b4f8865bab06cfa4f86fd94ebe31e921c76402428c94e15d448f8d0e0e",
    "04d65831029ae21c554c2ba442a55b92802a82840680b8b95cf4b939f7972f013883ba998139aae896aff41bb2eb73e1a2fb1bfaaf1958823b87a2b1830896981b",
    "048b86b0984fcc63434b12dc57f2c3896e7851cdb7ed907b18780ff75b47eff738355fc2b028dc2e16aedc63f3dd8d864fc450e141cc0f21b416676c693cf57ffd",
    "047870a256d9ac4e253c247dc71ad404c360f191cb40ecbb81004b6c8a80fbf610201bb57ded75874620a8aa1ea8cdcd6004e84242810fe4c03c0ac83dd42e570a",
    "04c6159283412ef6cfa74660565aa51d3fee4e7f852da62f9b33633f626cda4dfa8c39374940640b94fc075d1575545cc7685c969137b27e9d0d178feb2ca00623",
]


ATTESTATION_PUBKEYS_MAP: Dict[bytes, bytes] = {}
for pubkey_hex in ATTESTATION_PUBKEYS:
    pubkey_bytes = bytes.fromhex(pubkey_hex)
    ATTESTATION_PUBKEYS_MAP[hashlib.sha256(pubkey_bytes).digest()] = pubkey_bytes


OP_ATTESTATION = b"a"
OP_UNLOCK = b"u"
OP_I_CAN_HAS_HANDSHAEK = b"h"
OP_HER_COMEZ_TEH_HANDSHAEK = b"H"
OP_I_CAN_HAS_PAIRIN_VERIFICASHUN = b"v"
OP_NOISE_MSG = b"n"

RESPONSE_SUCCESS = b"\x00"
RESPONSE_FAILURE = b"\x01"

MIN_SUPPORTED_BITBOX02_MULTI_FIRMWARE_VERSION = semver.VersionInfo(9, 0, 0)
MIN_SUPPORTED_BITBOX02_BTCONLY_FIRMWARE_VERSION = semver.VersionInfo(9, 0, 0)
MIN_UNSUPPORTED_BITBOX02_MULTI_FIRMWARE_VERSION = semver.VersionInfo(10, 0, 0)
MIN_UNSUPPORTED_BITBOX02_BTCONLY_FIRMWARE_VERSION = semver.VersionInfo(10, 0, 0)


class Platform(enum.Enum):
    """Available hardware platforms"""

    BITBOX02 = "bitbox02"
    BITBOX02PLUS = "bitbox02-plus"


class BitBox02Edition(enum.Enum):
    """Editions for the BitBox02 platform"""

    MULTI = "multi"
    BTCONLY = "btconly"


class Bitbox02Exception(Exception):
    def __init__(self, code: int, message: str):
        self.code = code
        self.message = message
        super().__init__()

    def __str__(self) -> str:
        return f"error code: {self.code}, message: {self.message}"


class UserAbortException(Bitbox02Exception):
    pass


class AttestationException(Exception):
    pass


class FirmwareVersionOutdatedException(Exception):
    def __init__(self, version: semver.VersionInfo, required_version: semver.VersionInfo):
        super().__init__(
            "The BitBox02's firmware is not up to date. Device: {}, Required: {}".format(
                version, required_version
            )
        )


class LibraryVersionOutdatedException(Exception):
    def __init__(self, version: semver.VersionInfo):
        super().__init__(
            "The BitBox02's firmware version {} is too new for this app. Update the app".format(
                version
            )
        )


class UnsupportedException(Exception):
    def __init__(self, need_atleast: semver.VersionInfo):
        super().__init__(
            "This feature is supported from firmware version {}. Please upgrade your firmware.".format(
                need_atleast
            )
        )


class BitBoxNoiseConfig:
    """Stores Functions required setup a noise connection"""

    # pylint: disable=unused-argument
    def show_pairing(self, code: str, device_response: Callable[[], bool]) -> bool:
        """
        Returns True if the user confirms the pairing (both device and host).
        Returns False if the user rejects the pairing (either device or host).
        This function must call device_response() to invoke the pairing screen on the device.
        """
        return device_response()

    def attestation_check(self, result: bool) -> None:
        return

    def contains_device_static_pubkey(self, pubkey: bytes) -> bool:
        return False

    def add_device_static_pubkey(self, pubkey: bytes) -> None:
        pass

    def get_app_static_privkey(self) -> Optional[bytes]:
        return None

    def set_app_static_privkey(self, privkey: bytes) -> None:
        pass


class BitBoxProtocol(ABC):
    """
    Class for executing versioned BitBox operations
    (noise message transmissions, unlocks, etc).
    """

    def __init__(self, transport: TransportLayer):
        super().__init__()
        self._transport = transport
        self._noise: NoiseConnection = None

    def close(self) -> None:
        self._transport.close()

    def _raw_query(self, msg: bytes) -> bytes:
        cid = self._transport.generate_cid()
        return self._transport.query(msg, HWW_CMD, cid)

    def query(self, cmd: bytes, msg_data: bytes) -> Tuple[bytes, bytes]:
        """
        Encapsulates the given OP_* command and message data into a packet,
        and unpacks the response status code and data.
        """
        response = self._raw_query(cmd + msg_data)
        return response[:1], response[1:]

    @abstractmethod
    def _encode_noise_request(self, encrypted_msg: bytes) -> bytes:
        """Encapsulates an OP_NOISE_MSG message."""

    @abstractmethod
    def _decode_noise_response(self, encrypted_msg: bytes) -> Tuple[bytes, bytes]:
        """De-encapsulate an OP_NOISE_MSG response."""

    @abstractmethod
    def _handshake_query(self, req: bytes) -> Tuple[bytes, bytes]:
        """
        Executes a OP_HER_COMEZ_TEH_HANDSHAEK query with the given
        request data.
        Returns a pair (response status, response data).
        """

    def encrypted_query(self, msg: bytes) -> bytes:
        """
        Sends msg bytes and reads response bytes over an encrypted channel.
        """
        encrypted_msg = self._noise.encrypt(msg)
        encrypted_msg = self._encode_noise_request(encrypted_msg)

        response = self._raw_query(encrypted_msg)
        response_status, response = self._decode_noise_response(response)
        if response_status != RESPONSE_SUCCESS:
            raise Exception("Noise communication failed.")

        result = self._noise.decrypt(response)
        assert isinstance(result, bytes)
        return result

    # pylint: disable=too-many-branches
    def _create_noise_channel(self, noise_config: BitBoxNoiseConfig) -> NoiseConnection:
        if self._raw_query(OP_I_CAN_HAS_HANDSHAEK) != RESPONSE_SUCCESS:
            self.close()
            raise Exception("Couldn't kick off handshake")

        # init noise channel
        noise = NoiseConnection.from_name(b"Noise_XX_25519_ChaChaPoly_SHA256")
        noise.set_as_initiator()
        private_key = noise_config.get_app_static_privkey()
        if private_key is None:
            private_key = os.urandom(32)
            noise_config.set_app_static_privkey(private_key)
        noise.set_keypair_from_private_bytes(Keypair.STATIC, private_key)
        noise.set_prologue(b"Noise_XX_25519_ChaChaPoly_SHA256")
        noise.start_handshake()
        start_handshake_status, start_handshake_reply = self._handshake_query(noise.write_message())
        if start_handshake_status != RESPONSE_SUCCESS:
            self.close()
            raise Exception("Handshake process request failed.")
        noise.read_message(start_handshake_reply)
        remote_static_key = noise.noise_protocol.handshake_state.rs.public_bytes
        assert not noise.handshake_finished
        send_msg = noise.write_message()
        assert noise.handshake_finished
        pairing_code = base64.b32encode(noise.get_handshake_hash()).decode("ascii")
        handshake_finish_status, response = self._handshake_query(send_msg)
        if handshake_finish_status != RESPONSE_SUCCESS:
            self.close()
            raise Exception("Handshake conclusion failed.")

        # Check if we recognize the device's public key
        pairing_verification_required_by_host = True
        if noise_config.contains_device_static_pubkey(remote_static_key):
            pairing_verification_required_by_host = False

        pairing_verification_required_by_device = response == b"\x01"
        if pairing_verification_required_by_host or pairing_verification_required_by_device:

            def get_device_response() -> bool:
                device_response = self._raw_query(OP_I_CAN_HAS_PAIRIN_VERIFICASHUN)
                if device_response == RESPONSE_SUCCESS:
                    return True
                if device_response == RESPONSE_FAILURE:
                    return False
                raise Exception(f"Unexpected pairing response: f{repr(device_response)}")

            client_response_success = noise_config.show_pairing(
                "{} {}\n{} {}".format(
                    pairing_code[:5], pairing_code[5:10], pairing_code[10:15], pairing_code[15:20]
                ),
                get_device_response,
            )
            if not client_response_success:
                self.close()
                raise Exception("pairing rejected by the user")

            noise_config.add_device_static_pubkey(remote_static_key)
        return noise

    def noise_connect(self, noise_config: BitBoxNoiseConfig) -> None:
        self._noise = self._create_noise_channel(noise_config)

    @abstractmethod
    def unlock_query(self) -> None:
        """
        Executes an unlock query.
        Returns the bytes containing the response status.
        """

    @abstractmethod
    def cancel_outstanding_request(self) -> None:
        """
        Aborts/force close the outstanding request on the device.
        """


class BitBoxProtocolV1(BitBoxProtocol):
    """BitBox Protocol from firmware V1.0.0 onwards."""

    def unlock_query(self) -> None:
        raise NotImplementedError("unlock_query is not supported in BitBox protocol V1")

    def _encode_noise_request(self, encrypted_msg: bytes) -> bytes:
        return encrypted_msg

    def _decode_noise_response(self, encrypted_msg: bytes) -> Tuple[bytes, bytes]:
        """
        Until V7 of the protocol, we don't encapsulate OP_NOISE_MSG responses.
        Let's assume that if a response is empty, that means it
        contains an error.
        """
        if len(encrypted_msg) == 0:
            return RESPONSE_FAILURE, b""
        return RESPONSE_SUCCESS, encrypted_msg

    def _handshake_query(self, req: bytes) -> Tuple[bytes, bytes]:
        """
        V1-6 of the BB noise protocol doesn't encapsulate handshake requests, and don't
        send back a status code in the response.
        """
        noise_result = self._raw_query(req)
        return RESPONSE_SUCCESS, noise_result

    def cancel_outstanding_request(self) -> None:
        raise RuntimeError("cancel_outstanding_request should never be called here.")


class BitBoxProtocolV2(BitBoxProtocolV1):
    """BitBox Protocol from firmware V2.0.0 onwards."""

    def unlock_query(self) -> None:
        unlock_data = self._raw_query(OP_UNLOCK)
        if len(unlock_data) != 0:
            raise ValueError("OP_UNLOCK (V2) replied with wrong length.")


class BitBoxProtocolV3(BitBoxProtocolV2):
    """BitBox Protocol from firmware V3.0.0 onwards."""

    def unlock_query(self) -> None:
        unlock_result, unlock_data = self.query(OP_UNLOCK, b"")
        if len(unlock_data) != 0:
            raise ValueError("OP_UNLOCK (V3) replied with wrong length.")
        if unlock_result == RESPONSE_FAILURE:
            self.close()
            raise Exception("Unlock process aborted")


class BitBoxProtocolV4(BitBoxProtocolV3):
    """BitBox Protocol from firmware V4.0.0 onwards."""

    def _encode_noise_request(self, encrypted_msg: bytes) -> bytes:
        return OP_NOISE_MSG + encrypted_msg


class BitBoxProtocolV7(BitBoxProtocolV4):
    """Noise Protocol from firmware V7.0.0 onwards."""

    def __init__(self, transport: TransportLayer):
        super().__init__(transport)
        self.cancel_requested = False

    def _handshake_query(self, req: bytes) -> Tuple[bytes, bytes]:
        return self.query(OP_HER_COMEZ_TEH_HANDSHAEK, req)

    def _decode_noise_response(self, encrypted_msg: bytes) -> Tuple[bytes, bytes]:
        return encrypted_msg[:1], encrypted_msg[1:]

    def _raw_query(self, msg: bytes) -> bytes:
        """
        Starting with v7.0.0, HWW messages are encapsulated
        into an arbitration layer. The device can respond with
        RSP_NOT_READY to indicate that we should poll it later.
        """
        cid = self._transport.generate_cid()
        status = None
        payload: bytes
        while True:
            response = self._transport.query(HwwRequestCode.REQ_NEW + msg, HWW_CMD, cid)
            assert len(response) != 0, "Unexpected response of length 0 from HWW stack."
            status, payload = response[:1], response[1:]
            if status == HwwResponseCode.RSP_BUSY:
                assert (
                    len(payload) == 0
                ), "Unexpected payload of length {} with RSP_BUSY response.".format(len(payload))
                time.sleep(1)
            else:
                # We've successfully initiated our request.
                break

        if status in [HwwResponseCode.RSP_NACK]:
            # We should never receive a NACK unless some internal error occurs.
            raise Exception("Unexpected NACK response from HWW stack.")

        # The message has been sent. If we have a retry, poll the device until we're ready.
        self.cancel_requested = False
        while status == HwwResponseCode.RSP_NOT_READY:
            assert (
                len(payload) == 0
            ), "Unexpected payload of length {} with RSP_NOT_READY response.".format(len(payload))
            time.sleep(0.2)
            to_send = (
                HwwRequestCode.REQ_CANCEL if self.cancel_requested else HwwRequestCode.REQ_RETRY
            )
            response = self._transport.query(to_send, HWW_CMD, cid)
            assert len(response) != 0, "Unexpected response of length 0 from HWW stack."
            status, payload = response[:1], response[1:]
            if status not in [HwwResponseCode.RSP_NOT_READY, HwwResponseCode.RSP_ACK]:
                # We should never receive a NACK unless some internal error occurs.
                raise Exception(
                    "Unexpected response from HWW stack during retry ({}).".format(repr(status))
                )
        return payload

    def cancel_outstanding_request(self) -> None:
        self.cancel_requested = True


class BitBoxCommonAPI:
    """Class to communicate with a BitBox device"""

    # pylint: disable=too-many-public-methods,too-many-arguments
    def __init__(
        self,
        transport: TransportLayer,
        device_info: Optional[DeviceInfo],
        noise_config: BitBoxNoiseConfig,
    ):
        """
        Can raise LibraryVersionOutdatedException. check_min_version() should be called following
        the instantiation.
        If device_info is None, it is infered using the OP_INFO API call, available since
        firmware version v5.0.0.
        """
        self.debug = False

        if device_info is not None:
            version = device_info["serial_number"]
            if device_info["product_string"] in (BITBOX02MULTI, BITBOX02PLUS_MULTI):
                edition = BitBox02Edition.MULTI
            elif device_info["product_string"] in (BITBOX02BTC, BITBOX02PLUS_BTC):
                edition = BitBox02Edition.BTCONLY
            else:
                raise Exception("Invalid product string")
        else:
            version, _, edition, _, _ = self.get_info(transport)

        self.edition = edition
        try:
            self.version = parse_device_version(version)
        except:
            transport.close()
            raise

        # Delete the prelease part, as it messes with the comparison (e.g. 3.0.0-pre < 3.0.0 is
        # True, but the 3.0.0-pre has already the same API breaking changes like 3.0.0...).
        self.version = self.version.replace(prerelease=None)

        # raises exceptions if the library is out of date
        self._check_max_version()

        self._bitbox_protocol: BitBoxProtocol
        if self.version >= semver.VersionInfo(7, 0, 0):
            self._bitbox_protocol = BitBoxProtocolV7(transport)
        elif self.version >= semver.VersionInfo(4, 0, 0):
            self._bitbox_protocol = BitBoxProtocolV4(transport)
        elif self.version >= semver.VersionInfo(3, 0, 0):
            self._bitbox_protocol = BitBoxProtocolV3(transport)
        elif self.version >= semver.VersionInfo(2, 0, 0):
            self._bitbox_protocol = BitBoxProtocolV2(transport)
        else:
            self._bitbox_protocol = BitBoxProtocolV1(transport)

        if self.version >= semver.VersionInfo(2, 0, 0):
            noise_config.attestation_check(self._perform_attestation())
            self._bitbox_protocol.unlock_query()

        self._bitbox_protocol.noise_connect(noise_config)

    # pylint: disable=too-many-return-statements
    def _perform_attestation(self) -> bool:
        """Sends a random challenge and verifies that the response can be verified with
        Shift's root attestation pubkeys. Returns True if the verification is successful."""

        challenge = os.urandom(32)
        response_status, response = self._bitbox_protocol.query(OP_ATTESTATION, challenge)
        if response_status != RESPONSE_SUCCESS:
            return False

        # parse data
        bootloader_hash, response = response[:32], response[32:]
        device_pubkey_bytes, response = response[:64], response[64:]
        certificate, response = response[:64], response[64:]
        root_pubkey_identifier, response = response[:32], response[32:]
        challenge_signature, response = response[:64], response[64:]

        # check attestation
        if root_pubkey_identifier not in ATTESTATION_PUBKEYS_MAP:
            # root pubkey could not be identified.
            return False

        root_pubkey_bytes_uncompressed = ATTESTATION_PUBKEYS_MAP[root_pubkey_identifier]
        root_pubkey = ecdsa.VerifyingKey.from_string(
            root_pubkey_bytes_uncompressed[1:], ecdsa.curves.SECP256k1
        )

        device_pubkey = ecdsa.VerifyingKey.from_string(device_pubkey_bytes, ecdsa.curves.NIST256p)

        try:
            # Verify certificate
            if not root_pubkey.verify(
                certificate, bootloader_hash + device_pubkey_bytes, hashfunc=hashlib.sha256
            ):
                return False

            # Verify challenge
            if not device_pubkey.verify(challenge_signature, challenge, hashfunc=hashlib.sha256):
                return False
        except ecdsa.BadSignatureError:
            return False
        return True

    def _msg_query(
        self, request: hww.Request, expected_response: Optional[str] = None
    ) -> hww.Response:
        """
        Sends protobuf msg and retrieves protobuf response over an encrypted
        channel.
        """
        # pylint: disable=no-member
        if self.debug:
            print(request)
        response_bytes = self._bitbox_protocol.encrypted_query(request.SerializeToString())
        response = hww.Response()
        response.ParseFromString(response_bytes)
        if response.WhichOneof("response") == "error":
            if response.error.code == ERR_USER_ABORT:
                raise UserAbortException(response.error.code, response.error.message)
            raise Bitbox02Exception(response.error.code, response.error.message)
        if expected_response is not None and response.WhichOneof("response") != expected_response:
            raise Exception(
                "Unexpected response: {}, expected: {}".format(
                    response.WhichOneof("response"), expected_response
                )
            )
        if self.debug:
            print(response)
        return response

    def reboot(
        self, purpose: "system.RebootRequest.Purpose.V" = system.RebootRequest.Purpose.UPGRADE
    ) -> bool:
        """
        Sends the reboot request. If the user confirms the request on the device, the device reboots
        into the bootloader.
        The purpose defines what confirmation message the user gets to see on the device.
        """
        # pylint: disable=no-member
        request = hww.Request()
        request.reboot.CopyFrom(system.RebootRequest(purpose=purpose))
        try:
            self._msg_query(request)
        except OSError:
            # In case of reboot we can't read the response.
            return True
        except Bitbox02Exception:
            return False
        return True

    @staticmethod
    def get_info(
        transport: TransportLayer,
    ) -> Tuple[str, Platform, Union[BitBox02Edition], bool, Optional[bool]]:
        """
        Returns (version, platform, edition, unlocked, initialized).
        This is useful to get the version of the firmware or the device unlocked/initialized status
        when a usb descriptor is not available (via BitBoxBridge, etc.). The initialized status is
        supported from firmware v9.20.0 (it is None if not supported).
        This call does not use a versioned BitBoxProtocol for communication, as the version is not
        available (this call is used to get the version), so it must work for all firmware versions.
        """
        response = transport.query(HwwRequestCode.REQ_INFO, HWW_CMD, transport.generate_cid())

        version_str_len, response = int(response[0]), response[1:]
        version, response = response[:version_str_len], response[version_str_len:]
        version_str = version.rstrip(b"\0").decode("ascii")

        platform_byte, response = response[0], response[1:]
        platform = {0x00: Platform.BITBOX02, 0x02: Platform.BITBOX02PLUS}[platform_byte]

        edition_byte, response = response[0], response[1:]
        edition: Union[BitBox02Edition]
        if platform in (Platform.BITBOX02, Platform.BITBOX02PLUS):
            edition = {0x00: BitBox02Edition.MULTI, 0x01: BitBox02Edition.BTCONLY}[edition_byte]
        else:
            raise Exception("Unknown platform: {}".format(platform))

        unlocked_byte, response = response[0], response[1:]
        unlocked = {0x00: False, 0x01: True}[unlocked_byte]

        initialized = None
        if parse_device_version(version_str) >= semver.VersionInfo(9, 20, 0):
            initialized_byte = response[0]
            initialized = {0x00: False, 0x01: True}[initialized_byte]
        return (version_str, platform, edition, unlocked, initialized)

    def check_min_version(self) -> None:
        """
        Raises FirmwareVersionOutdatedException if the device has an older firmware version than
        required and the minimum required version.
        """
        if self.edition == BitBox02Edition.MULTI:
            if self.version < MIN_SUPPORTED_BITBOX02_MULTI_FIRMWARE_VERSION:
                raise FirmwareVersionOutdatedException(
                    self.version, MIN_SUPPORTED_BITBOX02_MULTI_FIRMWARE_VERSION
                )
        elif self.edition == BitBox02Edition.BTCONLY:
            if self.version < MIN_SUPPORTED_BITBOX02_BTCONLY_FIRMWARE_VERSION:
                raise FirmwareVersionOutdatedException(
                    self.version, MIN_SUPPORTED_BITBOX02_BTCONLY_FIRMWARE_VERSION
                )

    def cancel_outstanding_request(self) -> None:
        self._bitbox_protocol.cancel_outstanding_request()

    def _check_max_version(self) -> None:
        """
        Raises LibraryVersionOutdatedException if the device has an firmware which is too new
        (major version increased).
        """
        if self.edition == BitBox02Edition.MULTI:
            if self.version >= MIN_UNSUPPORTED_BITBOX02_MULTI_FIRMWARE_VERSION:
                raise LibraryVersionOutdatedException(self.version)
        elif self.edition == BitBox02Edition.BTCONLY:
            if self.version >= MIN_UNSUPPORTED_BITBOX02_BTCONLY_FIRMWARE_VERSION:
                raise LibraryVersionOutdatedException(self.version)

    def _require_atleast(self, version: semver.VersionInfo) -> None:
        """
        Raises UnsupportedException if the current firmware version is not at least the required version.
        """
        if self.version < version:
            raise UnsupportedException(version)

    def close(self) -> None:
        self._bitbox_protocol.close()



"""
Blockstream Jade Devices
************************
"""

from .jadepy import jade
from .jadepy.jade import JadeAPI, JadeError
from .jadepy.jade_serial import JadeSerialImpl

from serial.tools import list_ports

from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union
)
from ..descriptor import MultisigDescriptor
from ..hwwclient import HardwareWalletClient
from ..errors import (
    ActionCanceledError,
    BadArgumentError,
    DeviceConnectionError,
    DeviceFailureError,
    DeviceNotReadyError,
    UnavailableActionError,
    common_err_msgs,
    handle_errors
)
from ..common import (
    AddressType,
    Chain,
    sha256
)
from ..key import (
    ExtendedKey,
    KeyOriginInfo,
    is_hardened,
    parse_path
)
from ..psbt import PSBT
from .._script import (
    is_p2sh,
    is_p2wpkh,
    is_p2wsh,
    is_witness,
    parse_multisig
)

import base64
import logging
import semver
import os

# The test emulator port
SIMULATOR_PATH = 'tcp:127.0.0.1:30121'

HAS_NETWORKING = hasattr(jade, '_http_request')

py_enumerate = enumerate # To use the enumerate built-in, since the name is overridden below

def jade_exception(f: Callable[..., Any]) -> Any:
    @wraps(f)
    def func(*args: Any, **kwargs: Any) -> Any:
        try:
            return f(*args, **kwargs)
        except ValueError as e:
            raise BadArgumentError(str(e))
        except JadeError as e:
            if e.code == JadeError.USER_CANCELLED:
                raise ActionCanceledError(f'{f.__name__} canceled by user')
            elif e.code == JadeError.BAD_PARAMETERS:
                raise BadArgumentError(e.message)
            elif e.code == JadeError.INTERNAL_ERROR:
                raise DeviceFailureError(e.message)
            elif e.code == JadeError.HW_LOCKED:
                raise DeviceConnectionError('Device is locked')
            elif e.code == JadeError.NETWORK_MISMATCH:
                raise DeviceConnectionError('Network/chain selection error')
            elif e.code in [JadeError.INVALID_REQUEST, JadeError.UNKNOWN_METHOD, JadeError.PROTOCOL_ERROR]:
                raise DeviceConnectionError('Messaging/communiciation error')
            else:
                raise e
    return func

# This class extends the HardwareWalletClient for Blockstream Jade specific things
class JadeClient(HardwareWalletClient):
    MIN_SUPPORTED_FW_VERSION = semver.VersionInfo(0, 1, 32)
    PSBT_SUPPORTED_FW_VERSION = semver.VersionInfo(0, 1, 47)

    NETWORKS = {Chain.MAIN: 'mainnet',
                Chain.TEST: 'testnet',
                Chain.SIGNET: 'testnet',  # same as far as Jade is concerned
                Chain.REGTEST: 'localtest'}

    def _network(self) -> str:
        if self.chain not in self.NETWORKS:
            raise BadArgumentError(f'Unhandled network: {self.chain}')
        return self.NETWORKS[self.chain]

    ADDRTYPES = {AddressType.LEGACY: 'pkh(k)',
                 AddressType.WIT: 'wpkh(k)',
                 AddressType.SH_WIT: 'sh(wpkh(k))'}

    MULTI_ADDRTYPES = {AddressType.LEGACY: 'sh(multi(k))',
                       AddressType.WIT: 'wsh(multi(k))',
                       AddressType.SH_WIT: 'sh(wsh(multi(k)))'}

    @classmethod
    def _convertAddrType(cls, addrType: AddressType, multisig: bool) -> str:
        return cls.MULTI_ADDRTYPES[addrType] if multisig else cls.ADDRTYPES[addrType]

    # Derive a deterministic name for a multisig registration record (ignoring bip67 key sorting)
    @staticmethod
    def _get_multisig_name(type: str, threshold: int, signers: List[Tuple[bytes, Sequence[int]]]) -> str:
        # Concatenate script-type, threshold, and all signers fingerprints and derivation paths (sorted)
        summary = type + '|' + str(threshold) + '|'
        for fingerprint, path in sorted(signers):
            summary += fingerprint.hex() + '|' + str(path) + '|'

        # Hash it, get the first 6-bytes as hex, prepend with 'hwi'
        hash_summary = sha256(summary.encode()).hex()
        return 'hwi' + hash_summary[:12]

    def __init__(self, path: str, password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN, timeout: Optional[int] = None) -> None:
        super(JadeClient, self).__init__(path, password, expert, chain)
        self.jade = JadeAPI.create_serial(path, timeout=timeout)
        self.jade.connect()

        verinfo = self.jade.get_version_info()
        self.fw_version = semver.parse_version_info(verinfo['JADE_VERSION'])
        uninitialized = verinfo['JADE_STATE'] not in ['READY', 'TEMP']

        # Check minimum supported firmware version (ignore candidate/build parts)
        if self.MIN_SUPPORTED_FW_VERSION > self.fw_version.finalize_version():
            raise DeviceNotReadyError(f'Jade fw version: {self.fw_version} - minimum required version: {self.MIN_SUPPORTED_FW_VERSION}. '
                                      'Please update using a Blockstream Green companion app')
        if path == SIMULATOR_PATH:
            if uninitialized:
                # Connected to simulator but it appears to have no wallet set
                raise DeviceNotReadyError('Use JadeAPI.set_[seed|mnemonic] to set simulator wallet')
        else:
            if uninitialized and not HAS_NETWORKING:
                # Wallet not initialised/unlocked nor do we have networking dependencies
                # User must use 'Recovery Phrase Login' or 'QR Unlock' feature to access wallet
                raise DeviceNotReadyError('Use "Recovery Phrase Login" or "QR PIN Unlock" feature on Jade hw to access wallet')

            # Push some host entropy into jade
            self.jade.add_entropy(os.urandom(32))

            # Authenticate the user - this may require a PIN and pinserver interaction
            # (if we have required networking dependencies)
            authenticated = False
            while not authenticated:
                authenticated = self.jade.auth_user(self._network())

    # Retrieves the public key at the specified BIP 32 derivation path
    @jade_exception
    def get_pubkey_at_path(self, bip32_path: str) -> ExtendedKey:
        path = parse_path(bip32_path)
        xpub = self.jade.get_xpub(self._network(), path)
        ext_key = ExtendedKey.deserialize(xpub)
        return ext_key

    # Old firmware does not have native PSBT handling - walk the PSBT looking for inputs we can sign.
    # Push any signatures into the 'partial_sigs' map in the input, and return the updated PSBT.
    def legacy_sign_tx(self, tx: PSBT) -> PSBT:
        """
        Sign a transaction with the Blockstream Jade.
        """
        # Helper to get multisig record for change output
        def _parse_signers(hd_keypath_origins: List[KeyOriginInfo]) -> Tuple[List[Tuple[bytes, Sequence[int]]], List[Sequence[int]]]:
            # Split the path at the last hardened path element
            def _split_at_last_hardened_element(path: Sequence[int]) -> Tuple[Sequence[int], Sequence[int]]:
                for i in range(len(path), 0, -1):
                    if is_hardened(path[i - 1]):
                        return (path[:i], path[i:])
                return ([], path)

            signers = []
            paths = []
            for origin in hd_keypath_origins:
                prefix, suffix = _split_at_last_hardened_element(origin.path)
                signers.append((origin.fingerprint, prefix))
                paths.append(suffix)
            return signers, paths

        c_txn = tx.get_unsigned_tx()
        master_fp = self.get_master_fingerprint()
        signing_singlesigs = False
        signing_multisigs = {}
        need_to_sign = True

        while need_to_sign:
            signing_pubkeys: List[Optional[bytes]] = [None] * len(tx.inputs)
            need_to_sign = False

            # Signing input details
            jade_inputs = []
            for n_vin, psbtin in py_enumerate(tx.inputs):
                # Get bip32 path to use to sign, if required for this input
                path = None
                multisig_input = len(psbtin.hd_keypaths) > 1
                for pubkey, origin in psbtin.hd_keypaths.items():
                    if origin.fingerprint == master_fp and len(origin.path) > 0:
                        if not multisig_input:
                            signing_singlesigs = True

                        if psbtin.partial_sigs.get(pubkey, None) is None:
                            # hw to sign this input - it is not already signed
                            if signing_pubkeys[n_vin] is None:
                                signing_pubkeys[n_vin] = pubkey
                                path = origin.path
                            else:
                                # Additional signature needed for this input - ie. a multisig where this wallet is
                                # multiple signers?  Clumsy, but just loop and go through the signing procedure again.
                                need_to_sign = True

                # Get the tx and prevout/scriptcode
                utxo = None
                p2sh = False
                input_txn_bytes = None
                if psbtin.witness_utxo:
                    utxo = psbtin.witness_utxo
                if psbtin.non_witness_utxo:
                    if psbtin.prev_txid != psbtin.non_witness_utxo.hash:
                        raise BadArgumentError(f'Input {n_vin} has a non_witness_utxo with the wrong hash')
                    assert psbtin.prev_out is not None
                    utxo = psbtin.non_witness_utxo.vout[psbtin.prev_out]
                    input_txn_bytes = psbtin.non_witness_utxo.serialize_without_witness()
                if utxo is None:
                    raise Exception('PSBT is missing input utxo information, cannot sign')
                sats_value = utxo.nValue
                scriptcode = utxo.scriptPubKey

                if is_p2sh(scriptcode):
                    scriptcode = psbtin.redeem_script
                    p2sh = True

                witness_input, witness_version, witness_program = is_witness(scriptcode)

                if witness_input:
                    if is_p2wsh(scriptcode):
                        scriptcode = psbtin.witness_script
                    elif is_p2wpkh(scriptcode):
                        scriptcode = b'\x76\xa9\x14' + witness_program + b'\x88\xac'
                    else:
                        continue

                # If we are signing a multisig input, deduce the potential
                # registration details and cache as a potential change wallet
                if multisig_input and path and scriptcode and (p2sh or witness_input):
                    parsed = parse_multisig(scriptcode)
                    if parsed:
                        addr_type = AddressType.LEGACY if not witness_input else AddressType.WIT if not p2sh else AddressType.SH_WIT
                        script_variant = self._convertAddrType(addr_type, multisig=True)
                        threshold = parsed[0]

                        pubkeys = parsed[1]
                        hd_keypath_origins = [psbtin.hd_keypaths[pubkey] for pubkey in pubkeys]

                        signers, paths = _parse_signers(hd_keypath_origins)
                        multisig_name = self._get_multisig_name(script_variant, threshold, signers)
                        signing_multisigs[multisig_name] = (script_variant, threshold, signers)

                # Build the input and add to the list - include some host entropy for AE sigs (although we won't verify)
                jade_inputs.append({'is_witness': witness_input, 'satoshi': sats_value, 'script': scriptcode, 'path': path,
                                    'input_tx': input_txn_bytes, 'ae_host_entropy': os.urandom(32), 'ae_host_commitment': os.urandom(32)})

            # Change output details
            # This is optional, in that if we send it Jade validates the change output script
            # and the user need not confirm that output.  If not passed the change output must
            # be confirmed by the user on the hwwallet screen, like any other spend output.
            change: List[Optional[Dict[str, Any]]] = [None] * len(tx.outputs)

            # Skip automatic change validation in expert mode - user checks *every* output on hw
            if not self.expert:
                # If signing multisig inputs, get registered multisigs details in case we
                # see any multisig outputs which may be change which we can auto-validate.
                # ie. filter speculative 'signing multisigs' to ones actually registered on the hw
                if signing_multisigs:
                    registered_multisigs = self.jade.get_registered_multisigs()
                    signing_multisigs = {k: v for k, v in signing_multisigs.items()
                                         if k in registered_multisigs
                                         and registered_multisigs[k]['variant'] == v[0]
                                         and registered_multisigs[k]['threshold'] == v[1]
                                         and registered_multisigs[k]['num_signers'] == len(v[2])}

                # Look at every output...
                for n_vout, (txout, psbtout) in py_enumerate(zip(c_txn.vout, tx.outputs)):
                    num_signers = len(psbtout.hd_keypaths)

                    if num_signers == 1 and signing_singlesigs:
                        # Single-sig output - since we signed singlesig inputs this could be our change
                        for pubkey, origin in psbtout.hd_keypaths.items():
                            # Considers 'our' outputs as potential change as far as Jade is concerned
                            # ie. can be verified and auto-confirmed.
                            # Is this ok, or should check path also, assuming bip44-like ?
                            if origin.fingerprint == master_fp and len(origin.path) > 0:
                                change_addr_type = None
                                if txout.is_p2pkh():
                                    change_addr_type = AddressType.LEGACY
                                elif txout.is_witness()[0] and not txout.is_p2wsh():
                                    change_addr_type = AddressType.WIT  # ie. p2wpkh
                                elif txout.is_p2sh() and is_witness(psbtout.redeem_script)[0]:
                                    change_addr_type = AddressType.SH_WIT
                                else:
                                    continue

                                script_variant = self._convertAddrType(change_addr_type, multisig=False)
                                change[n_vout] = {'path': origin.path, 'variant': script_variant}

                    elif num_signers > 1 and signing_multisigs:
                        # Multisig output - since we signed multisig inputs this could be our change
                        candidate_multisigs = {k: v for k, v in signing_multisigs.items() if len(v[2]) == num_signers}
                        if not candidate_multisigs:
                            continue

                        for pubkey, origin in psbtout.hd_keypaths.items():
                            if origin.fingerprint == master_fp and len(origin.path) > 0:
                                change_addr_type = None
                                if txout.is_p2sh() and not is_witness(psbtout.redeem_script)[0]:
                                    change_addr_type = AddressType.LEGACY
                                    scriptcode = psbtout.redeem_script
                                elif txout.is_p2wsh() and not txout.is_p2sh():
                                    change_addr_type = AddressType.WIT
                                    scriptcode = psbtout.witness_script
                                elif txout.is_p2sh() and is_witness(psbtout.redeem_script)[0]:
                                    change_addr_type = AddressType.SH_WIT
                                    scriptcode = psbtout.witness_script
                                else:
                                    continue

                                parsed = parse_multisig(scriptcode)
                                if parsed:
                                    script_variant = self._convertAddrType(change_addr_type, multisig=True)
                                    threshold = parsed[0]

                                    pubkeys = parsed[1]
                                    hd_keypath_origins = [psbtout.hd_keypaths[pubkey] for pubkey in pubkeys]

                                    signers, paths = _parse_signers(hd_keypath_origins)
                                    multisig_name = self._get_multisig_name(script_variant, threshold, signers)
                                    matched_multisig = candidate_multisigs.get(multisig_name)

                                    if matched_multisig and matched_multisig[0] == script_variant and matched_multisig[1] == threshold and sorted(matched_multisig[2]) == sorted(signers):
                                        change[n_vout] = {'paths': paths, 'multisig_name': multisig_name}

            # The txn itself
            txn_bytes = c_txn.serialize_without_witness()

            # Request Jade generate the signatures for our inputs.
            # Change details are passed to be validated on the hw (user does not confirm)
            signatures = self.jade.sign_tx(self._network(), txn_bytes, jade_inputs, change, True)

            # Push sigs into PSBT structure as appropriate
            for psbtin, signer_pubkey, sigdata in zip(tx.inputs, signing_pubkeys, signatures):
                signer_commitment, sig = sigdata
                if signer_pubkey and sig:
                    psbtin.partial_sigs[signer_pubkey] = sig

        # Return the updated psbt
        return tx

    # Sign tx PSBT - newer Jade firmware supports native PSBT signing, but old firmwares require
    # mapping to the legacy 'sign_tx' structures.
    @jade_exception
    def sign_tx(self, tx: PSBT) -> PSBT:
        """
        Sign a transaction with the Blockstream Jade.
        """
        # Old firmware does not have native PSBT handling - use legacy method
        if self.PSBT_SUPPORTED_FW_VERSION > self.fw_version.finalize_version():
            return self.legacy_sign_tx(tx)

        # Firmware 0.1.47 (March 2023) and later support native PSBT signing
        psbt_b64 = tx.serialize()
        psbt_bytes = base64.b64decode(psbt_b64.strip())

        # NOTE: sign_psbt() does not use AE signatures, so sticks with default (rfc6979)
        psbt_bytes = self.jade.sign_psbt(self._network(), psbt_bytes)
        psbt_b64 = base64.b64encode(psbt_bytes).decode()

        psbt_signed = PSBT()
        psbt_signed.deserialize(psbt_b64)
        return psbt_signed

    # Sign message, confirmed on device
    @jade_exception
    def sign_message(self, message: Union[str, bytes], bip32_path: str) -> str:
        path = parse_path(bip32_path)
        if isinstance(message, bytes) or isinstance(message, bytearray):
            message = message.decode('utf-8')

        # NOTE: tests fail if we try to use AE signatures, so stick with default (rfc6979)
        signature = self.jade.sign_message(path, message)
        return str(signature)

    # Display address of specified type on the device.
    @jade_exception
    def display_singlesig_address(self, bip32_path: str, addr_type: AddressType) -> str:
        path = parse_path(bip32_path)
        script_variant = self._convertAddrType(addr_type, multisig=False)
        address = self.jade.get_receive_address(self._network(), path, variant=script_variant)
        return str(address)

    # Display multisig address of specified type on the device.
    @jade_exception
    def display_multisig_address(self, addr_type: AddressType, multisig: MultisigDescriptor) -> str:
        signer_origins = []
        signers = []
        paths = []
        for pubkey in multisig.pubkeys:
            if pubkey.extkey is None:
                raise BadArgumentError('Blockstream Jade can only generate addresses for multisigs with full extended keys')
            if pubkey.origin is None:
                raise BadArgumentError('Blockstream Jade can only generate addresses for multisigs with key origin information')
            if pubkey.deriv_path is None:
                raise BadArgumentError('Blockstream Jade can only generate addresses for multisigs with key derivation paths')

            if pubkey.origin.path and not is_hardened(pubkey.origin.path[-1]):
                logging.warning(f'Final element of origin path {pubkey.origin.path} unhardened')
                logging.warning('Blockstream Jade may not be able to identify change sent back to this descriptor')

            # Tuple to derive deterministic name for the registrtion
            signer_origins.append((pubkey.origin.fingerprint, pubkey.origin.path))

            #  We won't include the additional path in the multisig registration
            signers.append({'fingerprint': pubkey.origin.fingerprint,
                            'derivation': pubkey.origin.path,
                            'xpub': pubkey.pubkey,
                            'path': []})

            # Instead hold it as the address path
            path = pubkey.deriv_path[1:] if pubkey.deriv_path[0] == '/' else pubkey.deriv_path
            paths.append(parse_path(path))

        if multisig.is_sorted and paths[:-1] != paths[1:]:
            logging.warning('Sorted multisig with different derivations per signer')
            logging.warning('Blockstream Jade may not be able to validate change sent back to this descriptor')

        # Get a deterministic name for this multisig wallet (ignoring bip67 key sorting)
        script_variant = self._convertAddrType(addr_type, multisig=True)
        multisig_name = self._get_multisig_name(script_variant, multisig.thresh, signer_origins)

        # Need to ensure this multisig wallet is registered first
        # (Note: 're-registering' is a no-op)
        self.jade.register_multisig(self._network(), multisig_name, script_variant, multisig.is_sorted, multisig.thresh, signers)
        address = self.jade.get_receive_address(self._network(), paths, multisig_name=multisig_name)

        return str(address)

    # Setup a new device
    def setup_device(self, label: str = "", passphrase: str = "") -> bool:
        """
        Blockstream Jade does not support setup via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('Blockstream Jade does not support software setup')

    # Wipe this device
    def wipe_device(self) -> bool:
        """
        Blockstream Jade does not support wiping via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('Blockstream Jade does not support wiping via software')

    # Restore device from mnemonic or xprv
    def restore_device(self, label: str = "", word_count: int = 24) -> bool:
        """
        Blockstream Jade does not support restoring via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('Blockstream Jade does not support restoring via software')

    # Begin backup process
    def backup_device(self, label: str = "", passphrase: str = "") -> bool:
        """
        Blockstream Jade does not support backing up via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('Blockstream Jade does not support creating a backup via software')

    # Close the device
    def close(self) -> None:
        self.jade.disconnect()

    # Prompt pin
    def prompt_pin(self) -> bool:
        """
        Blockstream Jade does not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('Blockstream Jade does not need a PIN sent from the host')

    # Send pin
    def send_pin(self, pin: str) -> bool:
        """
        Blockstream Jade does not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('Blockstream Jade does not need a PIN sent from the host')

    # Toggle passphrase
    def toggle_passphrase(self) -> bool:
        """
        Blockstream Jade does not support toggling passphrase from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError('Blockstream Jade does not support toggling passphrase from the host')

    @jade_exception
    def can_sign_taproot(self) -> bool:
        """
        Blockstream Jade does not currently support Taproot.

        :returns: False, always
        """
        return False


def enumerate(password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN, allow_emulators: bool = False) -> List[Dict[str, Any]]:
    results = []

    def _get_device_entry(device_model: str, device_path: str) -> Dict[str, Any]:
        d_data: Dict[str, Any] = {}
        d_data['type'] = 'jade'
        d_data['model'] = device_model
        d_data['path'] = device_path
        d_data['needs_pin_sent'] = False
        d_data['needs_passphrase_sent'] = False

        client = None
        with handle_errors(common_err_msgs['enumerate'], d_data):
            client = JadeClient(device_path, password, expert, chain, timeout=1)
            d_data['fingerprint'] = client.get_master_fingerprint().hex()

        if client:
            client.close()

        return d_data

    # Jade is not really an HID device, it shows as a serial/com port device.
    # Scan com ports looking for the relevant vid and pid, and use 'path' to
    # hold the path to the serial port device, eg. /dev/ttyUSB0
    for devinfo in list_ports.comports():
        if (devinfo.vid, devinfo.pid) in JadeSerialImpl.JADE_DEVICE_IDS:
            results.append(_get_device_entry('jade', devinfo.device))

    # If we can connect to the simulator, add it too
    if allow_emulators:
        try:
            with JadeAPI.create_serial(SIMULATOR_PATH, timeout=1) as jade:
                verinfo = jade.get_version_info()

            if verinfo is not None:
                results.append(_get_device_entry('jade_simulator', SIMULATOR_PATH))

        except Exception as e:
            # If we get any sort of error do not add the simulator
            logging.debug(f'Failed to connect to Jade simulator at {SIMULATOR_PATH}')
            logging.debug(e)

    return results



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2019 Shift Cryptosecurity AG

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import antiklepto_pb2
import builtins
import collections.abc
from . import common_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _ETHCoin:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ETHCoinEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ETHCoin.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    ETH: _ETHCoin.ValueType  # 0
    RopstenETH: _ETHCoin.ValueType  # 1
    """Removed in v9.14.0 - deprecated"""
    RinkebyETH: _ETHCoin.ValueType  # 2
    """Removed in v9.14.0 - deprecated"""

class ETHCoin(_ETHCoin, metaclass=_ETHCoinEnumTypeWrapper):
    """Kept for backwards compatibility. Use chain_id instead, introduced in v9.10.0."""

ETH: ETHCoin.ValueType  # 0
RopstenETH: ETHCoin.ValueType  # 1
"""Removed in v9.14.0 - deprecated"""
RinkebyETH: ETHCoin.ValueType  # 2
"""Removed in v9.14.0 - deprecated"""
global___ETHCoin = ETHCoin

class _ETHAddressCase:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ETHAddressCaseEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_ETHAddressCase.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    ETH_ADDRESS_CASE_MIXED: _ETHAddressCase.ValueType  # 0
    ETH_ADDRESS_CASE_UPPER: _ETHAddressCase.ValueType  # 1
    ETH_ADDRESS_CASE_LOWER: _ETHAddressCase.ValueType  # 2

class ETHAddressCase(_ETHAddressCase, metaclass=_ETHAddressCaseEnumTypeWrapper): ...

ETH_ADDRESS_CASE_MIXED: ETHAddressCase.ValueType  # 0
ETH_ADDRESS_CASE_UPPER: ETHAddressCase.ValueType  # 1
ETH_ADDRESS_CASE_LOWER: ETHAddressCase.ValueType  # 2
global___ETHAddressCase = ETHAddressCase

@typing.final
class ETHPubRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _OutputType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _OutputTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ETHPubRequest._OutputType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        ADDRESS: ETHPubRequest._OutputType.ValueType  # 0
        XPUB: ETHPubRequest._OutputType.ValueType  # 1

    class OutputType(_OutputType, metaclass=_OutputTypeEnumTypeWrapper): ...
    ADDRESS: ETHPubRequest.OutputType.ValueType  # 0
    XPUB: ETHPubRequest.OutputType.ValueType  # 1

    KEYPATH_FIELD_NUMBER: builtins.int
    COIN_FIELD_NUMBER: builtins.int
    OUTPUT_TYPE_FIELD_NUMBER: builtins.int
    DISPLAY_FIELD_NUMBER: builtins.int
    CONTRACT_ADDRESS_FIELD_NUMBER: builtins.int
    CHAIN_ID_FIELD_NUMBER: builtins.int
    coin: global___ETHCoin.ValueType
    """Deprecated: use chain_id instead."""
    output_type: global___ETHPubRequest.OutputType.ValueType
    display: builtins.bool
    contract_address: builtins.bytes
    chain_id: builtins.int
    """If non-zero, `coin` is ignored and `chain_id` is used to identify the network."""
    @property
    def keypath(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        keypath: collections.abc.Iterable[builtins.int] | None = ...,
        coin: global___ETHCoin.ValueType = ...,
        output_type: global___ETHPubRequest.OutputType.ValueType = ...,
        display: builtins.bool = ...,
        contract_address: builtins.bytes = ...,
        chain_id: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["chain_id", b"chain_id", "coin", b"coin", "contract_address", b"contract_address", "display", b"display", "keypath", b"keypath", "output_type", b"output_type"]) -> None: ...

global___ETHPubRequest = ETHPubRequest

@typing.final
class ETHSignRequest(google.protobuf.message.Message):
    """TX payload for "legacy" (EIP-155) transactions: https://eips.ethereum.org/EIPS/eip-155"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COIN_FIELD_NUMBER: builtins.int
    KEYPATH_FIELD_NUMBER: builtins.int
    NONCE_FIELD_NUMBER: builtins.int
    GAS_PRICE_FIELD_NUMBER: builtins.int
    GAS_LIMIT_FIELD_NUMBER: builtins.int
    RECIPIENT_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    HOST_NONCE_COMMITMENT_FIELD_NUMBER: builtins.int
    CHAIN_ID_FIELD_NUMBER: builtins.int
    ADDRESS_CASE_FIELD_NUMBER: builtins.int
    coin: global___ETHCoin.ValueType
    """Deprecated: use chain_id instead."""
    nonce: builtins.bytes
    """smallest big endian serialization, max. 16 bytes"""
    gas_price: builtins.bytes
    """smallest big endian serialization, max. 16 bytes"""
    gas_limit: builtins.bytes
    """smallest big endian serialization, max. 16 bytes"""
    recipient: builtins.bytes
    """20 byte recipient"""
    value: builtins.bytes
    """smallest big endian serialization, max. 32 bytes"""
    data: builtins.bytes
    chain_id: builtins.int
    """If non-zero, `coin` is ignored and `chain_id` is used to identify the network."""
    address_case: global___ETHAddressCase.ValueType
    @property
    def keypath(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def host_nonce_commitment(self) -> antiklepto_pb2.AntiKleptoHostNonceCommitment: ...
    def __init__(
        self,
        *,
        coin: global___ETHCoin.ValueType = ...,
        keypath: collections.abc.Iterable[builtins.int] | None = ...,
        nonce: builtins.bytes = ...,
        gas_price: builtins.bytes = ...,
        gas_limit: builtins.bytes = ...,
        recipient: builtins.bytes = ...,
        value: builtins.bytes = ...,
        data: builtins.bytes = ...,
        host_nonce_commitment: antiklepto_pb2.AntiKleptoHostNonceCommitment | None = ...,
        chain_id: builtins.int = ...,
        address_case: global___ETHAddressCase.ValueType = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["host_nonce_commitment", b"host_nonce_commitment"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["address_case", b"address_case", "chain_id", b"chain_id", "coin", b"coin", "data", b"data", "gas_limit", b"gas_limit", "gas_price", b"gas_price", "host_nonce_commitment", b"host_nonce_commitment", "keypath", b"keypath", "nonce", b"nonce", "recipient", b"recipient", "value", b"value"]) -> None: ...

global___ETHSignRequest = ETHSignRequest

@typing.final
class ETHSignEIP1559Request(google.protobuf.message.Message):
    """TX payload for an EIP-1559 (type 2) transaction: https://eips.ethereum.org/EIPS/eip-1559"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CHAIN_ID_FIELD_NUMBER: builtins.int
    KEYPATH_FIELD_NUMBER: builtins.int
    NONCE_FIELD_NUMBER: builtins.int
    MAX_PRIORITY_FEE_PER_GAS_FIELD_NUMBER: builtins.int
    MAX_FEE_PER_GAS_FIELD_NUMBER: builtins.int
    GAS_LIMIT_FIELD_NUMBER: builtins.int
    RECIPIENT_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    HOST_NONCE_COMMITMENT_FIELD_NUMBER: builtins.int
    ADDRESS_CASE_FIELD_NUMBER: builtins.int
    chain_id: builtins.int
    nonce: builtins.bytes
    """smallest big endian serialization, max. 16 bytes"""
    max_priority_fee_per_gas: builtins.bytes
    """smallest big endian serialization, max. 16 bytes"""
    max_fee_per_gas: builtins.bytes
    """smallest big endian serialization, max. 16 bytes"""
    gas_limit: builtins.bytes
    """smallest big endian serialization, max. 16 bytes"""
    recipient: builtins.bytes
    """20 byte recipient"""
    value: builtins.bytes
    """smallest big endian serialization, max. 32 bytes"""
    data: builtins.bytes
    address_case: global___ETHAddressCase.ValueType
    @property
    def keypath(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def host_nonce_commitment(self) -> antiklepto_pb2.AntiKleptoHostNonceCommitment: ...
    def __init__(
        self,
        *,
        chain_id: builtins.int = ...,
        keypath: collections.abc.Iterable[builtins.int] | None = ...,
        nonce: builtins.bytes = ...,
        max_priority_fee_per_gas: builtins.bytes = ...,
        max_fee_per_gas: builtins.bytes = ...,
        gas_limit: builtins.bytes = ...,
        recipient: builtins.bytes = ...,
        value: builtins.bytes = ...,
        data: builtins.bytes = ...,
        host_nonce_commitment: antiklepto_pb2.AntiKleptoHostNonceCommitment | None = ...,
        address_case: global___ETHAddressCase.ValueType = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["host_nonce_commitment", b"host_nonce_commitment"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["address_case", b"address_case", "chain_id", b"chain_id", "data", b"data", "gas_limit", b"gas_limit", "host_nonce_commitment", b"host_nonce_commitment", "keypath", b"keypath", "max_fee_per_gas", b"max_fee_per_gas", "max_priority_fee_per_gas", b"max_priority_fee_per_gas", "nonce", b"nonce", "recipient", b"recipient", "value", b"value"]) -> None: ...

global___ETHSignEIP1559Request = ETHSignEIP1559Request

@typing.final
class ETHSignMessageRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    COIN_FIELD_NUMBER: builtins.int
    KEYPATH_FIELD_NUMBER: builtins.int
    MSG_FIELD_NUMBER: builtins.int
    HOST_NONCE_COMMITMENT_FIELD_NUMBER: builtins.int
    CHAIN_ID_FIELD_NUMBER: builtins.int
    coin: global___ETHCoin.ValueType
    """Deprecated: use chain_id instead."""
    msg: builtins.bytes
    chain_id: builtins.int
    """If non-zero, `coin` is ignored and `chain_id` is used to identify the network."""
    @property
    def keypath(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def host_nonce_commitment(self) -> antiklepto_pb2.AntiKleptoHostNonceCommitment: ...
    def __init__(
        self,
        *,
        coin: global___ETHCoin.ValueType = ...,
        keypath: collections.abc.Iterable[builtins.int] | None = ...,
        msg: builtins.bytes = ...,
        host_nonce_commitment: antiklepto_pb2.AntiKleptoHostNonceCommitment | None = ...,
        chain_id: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["host_nonce_commitment", b"host_nonce_commitment"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["chain_id", b"chain_id", "coin", b"coin", "host_nonce_commitment", b"host_nonce_commitment", "keypath", b"keypath", "msg", b"msg"]) -> None: ...

global___ETHSignMessageRequest = ETHSignMessageRequest

@typing.final
class ETHSignResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SIGNATURE_FIELD_NUMBER: builtins.int
    signature: builtins.bytes
    """65 bytes, last byte is the recid"""
    def __init__(
        self,
        *,
        signature: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["signature", b"signature"]) -> None: ...

global___ETHSignResponse = ETHSignResponse

@typing.final
class ETHSignTypedMessageRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _DataType:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _DataTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ETHSignTypedMessageRequest._DataType.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN: ETHSignTypedMessageRequest._DataType.ValueType  # 0
        BYTES: ETHSignTypedMessageRequest._DataType.ValueType  # 1
        UINT: ETHSignTypedMessageRequest._DataType.ValueType  # 2
        INT: ETHSignTypedMessageRequest._DataType.ValueType  # 3
        BOOL: ETHSignTypedMessageRequest._DataType.ValueType  # 4
        ADDRESS: ETHSignTypedMessageRequest._DataType.ValueType  # 5
        STRING: ETHSignTypedMessageRequest._DataType.ValueType  # 6
        ARRAY: ETHSignTypedMessageRequest._DataType.ValueType  # 7
        STRUCT: ETHSignTypedMessageRequest._DataType.ValueType  # 8

    class DataType(_DataType, metaclass=_DataTypeEnumTypeWrapper): ...
    UNKNOWN: ETHSignTypedMessageRequest.DataType.ValueType  # 0
    BYTES: ETHSignTypedMessageRequest.DataType.ValueType  # 1
    UINT: ETHSignTypedMessageRequest.DataType.ValueType  # 2
    INT: ETHSignTypedMessageRequest.DataType.ValueType  # 3
    BOOL: ETHSignTypedMessageRequest.DataType.ValueType  # 4
    ADDRESS: ETHSignTypedMessageRequest.DataType.ValueType  # 5
    STRING: ETHSignTypedMessageRequest.DataType.ValueType  # 6
    ARRAY: ETHSignTypedMessageRequest.DataType.ValueType  # 7
    STRUCT: ETHSignTypedMessageRequest.DataType.ValueType  # 8

    @typing.final
    class MemberType(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        TYPE_FIELD_NUMBER: builtins.int
        SIZE_FIELD_NUMBER: builtins.int
        STRUCT_NAME_FIELD_NUMBER: builtins.int
        ARRAY_TYPE_FIELD_NUMBER: builtins.int
        type: global___ETHSignTypedMessageRequest.DataType.ValueType
        size: builtins.int
        struct_name: builtins.str
        """if type==STRUCT, name of struct type."""
        @property
        def array_type(self) -> global___ETHSignTypedMessageRequest.MemberType:
            """if type==ARRAY, type of elements"""

        def __init__(
            self,
            *,
            type: global___ETHSignTypedMessageRequest.DataType.ValueType = ...,
            size: builtins.int = ...,
            struct_name: builtins.str = ...,
            array_type: global___ETHSignTypedMessageRequest.MemberType | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["array_type", b"array_type"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["array_type", b"array_type", "size", b"size", "struct_name", b"struct_name", "type", b"type"]) -> None: ...

    @typing.final
    class Member(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        TYPE_FIELD_NUMBER: builtins.int
        name: builtins.str
        @property
        def type(self) -> global___ETHSignTypedMessageRequest.MemberType: ...
        def __init__(
            self,
            *,
            name: builtins.str = ...,
            type: global___ETHSignTypedMessageRequest.MemberType | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["type", b"type"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["name", b"name", "type", b"type"]) -> None: ...

    @typing.final
    class StructType(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        MEMBERS_FIELD_NUMBER: builtins.int
        name: builtins.str
        @property
        def members(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ETHSignTypedMessageRequest.Member]: ...
        def __init__(
            self,
            *,
            name: builtins.str = ...,
            members: collections.abc.Iterable[global___ETHSignTypedMessageRequest.Member] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["members", b"members", "name", b"name"]) -> None: ...

    CHAIN_ID_FIELD_NUMBER: builtins.int
    KEYPATH_FIELD_NUMBER: builtins.int
    TYPES_FIELD_NUMBER: builtins.int
    PRIMARY_TYPE_FIELD_NUMBER: builtins.int
    HOST_NONCE_COMMITMENT_FIELD_NUMBER: builtins.int
    chain_id: builtins.int
    primary_type: builtins.str
    @property
    def keypath(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def types(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ETHSignTypedMessageRequest.StructType]: ...
    @property
    def host_nonce_commitment(self) -> antiklepto_pb2.AntiKleptoHostNonceCommitment: ...
    def __init__(
        self,
        *,
        chain_id: builtins.int = ...,
        keypath: collections.abc.Iterable[builtins.int] | None = ...,
        types: collections.abc.Iterable[global___ETHSignTypedMessageRequest.StructType] | None = ...,
        primary_type: builtins.str = ...,
        host_nonce_commitment: antiklepto_pb2.AntiKleptoHostNonceCommitment | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["host_nonce_commitment", b"host_nonce_commitment"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["chain_id", b"chain_id", "host_nonce_commitment", b"host_nonce_commitment", "keypath", b"keypath", "primary_type", b"primary_type", "types", b"types"]) -> None: ...

global___ETHSignTypedMessageRequest = ETHSignTypedMessageRequest

@typing.final
class ETHTypedMessageValueResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _RootObject:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _RootObjectEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[ETHTypedMessageValueResponse._RootObject.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN: ETHTypedMessageValueResponse._RootObject.ValueType  # 0
        DOMAIN: ETHTypedMessageValueResponse._RootObject.ValueType  # 1
        MESSAGE: ETHTypedMessageValueResponse._RootObject.ValueType  # 2

    class RootObject(_RootObject, metaclass=_RootObjectEnumTypeWrapper): ...
    UNKNOWN: ETHTypedMessageValueResponse.RootObject.ValueType  # 0
    DOMAIN: ETHTypedMessageValueResponse.RootObject.ValueType  # 1
    MESSAGE: ETHTypedMessageValueResponse.RootObject.ValueType  # 2

    ROOT_OBJECT_FIELD_NUMBER: builtins.int
    PATH_FIELD_NUMBER: builtins.int
    root_object: global___ETHTypedMessageValueResponse.RootObject.ValueType
    @property
    def path(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        root_object: global___ETHTypedMessageValueResponse.RootObject.ValueType = ...,
        path: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["path", b"path", "root_object", b"root_object"]) -> None: ...

global___ETHTypedMessageValueResponse = ETHTypedMessageValueResponse

@typing.final
class ETHTypedMessageValueRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    value: builtins.bytes
    def __init__(
        self,
        *,
        value: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["value", b"value"]) -> None: ...

global___ETHTypedMessageValueRequest = ETHTypedMessageValueRequest

@typing.final
class ETHRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PUB_FIELD_NUMBER: builtins.int
    SIGN_FIELD_NUMBER: builtins.int
    SIGN_MSG_FIELD_NUMBER: builtins.int
    ANTIKLEPTO_SIGNATURE_FIELD_NUMBER: builtins.int
    SIGN_TYPED_MSG_FIELD_NUMBER: builtins.int
    TYPED_MSG_VALUE_FIELD_NUMBER: builtins.int
    SIGN_EIP1559_FIELD_NUMBER: builtins.int
    @property
    def pub(self) -> global___ETHPubRequest: ...
    @property
    def sign(self) -> global___ETHSignRequest: ...
    @property
    def sign_msg(self) -> global___ETHSignMessageRequest: ...
    @property
    def antiklepto_signature(self) -> antiklepto_pb2.AntiKleptoSignatureRequest: ...
    @property
    def sign_typed_msg(self) -> global___ETHSignTypedMessageRequest: ...
    @property
    def typed_msg_value(self) -> global___ETHTypedMessageValueRequest: ...
    @property
    def sign_eip1559(self) -> global___ETHSignEIP1559Request: ...
    def __init__(
        self,
        *,
        pub: global___ETHPubRequest | None = ...,
        sign: global___ETHSignRequest | None = ...,
        sign_msg: global___ETHSignMessageRequest | None = ...,
        antiklepto_signature: antiklepto_pb2.AntiKleptoSignatureRequest | None = ...,
        sign_typed_msg: global___ETHSignTypedMessageRequest | None = ...,
        typed_msg_value: global___ETHTypedMessageValueRequest | None = ...,
        sign_eip1559: global___ETHSignEIP1559Request | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["antiklepto_signature", b"antiklepto_signature", "pub", b"pub", "request", b"request", "sign", b"sign", "sign_eip1559", b"sign_eip1559", "sign_msg", b"sign_msg", "sign_typed_msg", b"sign_typed_msg", "typed_msg_value", b"typed_msg_value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["antiklepto_signature", b"antiklepto_signature", "pub", b"pub", "request", b"request", "sign", b"sign", "sign_eip1559", b"sign_eip1559", "sign_msg", b"sign_msg", "sign_typed_msg", b"sign_typed_msg", "typed_msg_value", b"typed_msg_value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["request", b"request"]) -> typing.Literal["pub", "sign", "sign_msg", "antiklepto_signature", "sign_typed_msg", "typed_msg_value", "sign_eip1559"] | None: ...

global___ETHRequest = ETHRequest

@typing.final
class ETHResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PUB_FIELD_NUMBER: builtins.int
    SIGN_FIELD_NUMBER: builtins.int
    ANTIKLEPTO_SIGNER_COMMITMENT_FIELD_NUMBER: builtins.int
    TYPED_MSG_VALUE_FIELD_NUMBER: builtins.int
    @property
    def pub(self) -> common_pb2.PubResponse: ...
    @property
    def sign(self) -> global___ETHSignResponse: ...
    @property
    def antiklepto_signer_commitment(self) -> antiklepto_pb2.AntiKleptoSignerCommitment: ...
    @property
    def typed_msg_value(self) -> global___ETHTypedMessageValueResponse: ...
    def __init__(
        self,
        *,
        pub: common_pb2.PubResponse | None = ...,
        sign: global___ETHSignResponse | None = ...,
        antiklepto_signer_commitment: antiklepto_pb2.AntiKleptoSignerCommitment | None = ...,
        typed_msg_value: global___ETHTypedMessageValueResponse | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["antiklepto_signer_commitment", b"antiklepto_signer_commitment", "pub", b"pub", "response", b"response", "sign", b"sign", "typed_msg_value", b"typed_msg_value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["antiklepto_signer_commitment", b"antiklepto_signer_commitment", "pub", b"pub", "response", b"response", "sign", b"sign", "typed_msg_value", b"typed_msg_value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["response", b"response"]) -> typing.Literal["pub", "sign", "antiklepto_signer_commitment", "typed_msg_value"] | None: ...

global___ETHResponse = ETHResponse



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2019 Shift Cryptosecurity AG

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class RebootRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Purpose:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _PurposeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[RebootRequest._Purpose.ValueType], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UPGRADE: RebootRequest._Purpose.ValueType  # 0
        SETTINGS: RebootRequest._Purpose.ValueType  # 1

    class Purpose(_Purpose, metaclass=_PurposeEnumTypeWrapper): ...
    UPGRADE: RebootRequest.Purpose.ValueType  # 0
    SETTINGS: RebootRequest.Purpose.ValueType  # 1

    PURPOSE_FIELD_NUMBER: builtins.int
    purpose: global___RebootRequest.Purpose.ValueType
    def __init__(
        self,
        *,
        purpose: global___RebootRequest.Purpose.ValueType = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["purpose", b"purpose"]) -> None: ...

global___RebootRequest = RebootRequest



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cardano.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import common_pb2 as common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\rcardano.proto\x12\x14shiftcrypto.bitbox02\x1a\x0c\x63ommon.proto\"F\n\x13\x43\x61rdanoXpubsRequest\x12/\n\x08keypaths\x18\x01 \x03(\x0b\x32\x1d.shiftcrypto.bitbox02.Keypath\"%\n\x14\x43\x61rdanoXpubsResponse\x12\r\n\x05xpubs\x18\x01 \x03(\x0c\"\x9e\x01\n\x13\x43\x61rdanoScriptConfig\x12\x43\n\x07pkh_skh\x18\x01 \x01(\x0b\x32\x30.shiftcrypto.bitbox02.CardanoScriptConfig.PkhSkhH\x00\x1a\x38\n\x06PkhSkh\x12\x17\n\x0fkeypath_payment\x18\x01 \x03(\r\x12\x15\n\rkeypath_stake\x18\x02 \x03(\rB\x08\n\x06\x63onfig\"\xa1\x01\n\x15\x43\x61rdanoAddressRequest\x12\x35\n\x07network\x18\x01 \x01(\x0e\x32$.shiftcrypto.bitbox02.CardanoNetwork\x12\x0f\n\x07\x64isplay\x18\x02 \x01(\x08\x12@\n\rscript_config\x18\x03 \x01(\x0b\x32).shiftcrypto.bitbox02.CardanoScriptConfig\"\xb0\r\n\x1d\x43\x61rdanoSignTransactionRequest\x12\x35\n\x07network\x18\x01 \x01(\x0e\x32$.shiftcrypto.bitbox02.CardanoNetwork\x12I\n\x06inputs\x18\x02 \x03(\x0b\x32\x39.shiftcrypto.bitbox02.CardanoSignTransactionRequest.Input\x12K\n\x07outputs\x18\x03 \x03(\x0b\x32:.shiftcrypto.bitbox02.CardanoSignTransactionRequest.Output\x12\x0b\n\x03\x66\x65\x65\x18\x04 \x01(\x04\x12\x0b\n\x03ttl\x18\x05 \x01(\x04\x12U\n\x0c\x63\x65rtificates\x18\x06 \x03(\x0b\x32?.shiftcrypto.bitbox02.CardanoSignTransactionRequest.Certificate\x12S\n\x0bwithdrawals\x18\x07 \x03(\x0b\x32>.shiftcrypto.bitbox02.CardanoSignTransactionRequest.Withdrawal\x12\x1f\n\x17validity_interval_start\x18\x08 \x01(\x04\x12\x16\n\x0e\x61llow_zero_ttl\x18\t \x01(\x08\x12\x15\n\rtag_cbor_sets\x18\n \x01(\x08\x1aG\n\x05Input\x12\x0f\n\x07keypath\x18\x01 \x03(\r\x12\x15\n\rprev_out_hash\x18\x02 \x01(\x0c\x12\x16\n\x0eprev_out_index\x18\x03 \x01(\r\x1a\xa1\x01\n\nAssetGroup\x12\x11\n\tpolicy_id\x18\x01 \x01(\x0c\x12T\n\x06tokens\x18\x02 \x03(\x0b\x32\x44.shiftcrypto.bitbox02.CardanoSignTransactionRequest.AssetGroup.Token\x1a*\n\x05Token\x12\x12\n\nasset_name\x18\x01 \x01(\x0c\x12\r\n\x05value\x18\x02 \x01(\x04\x1a\xc8\x01\n\x06Output\x12\x17\n\x0f\x65ncoded_address\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04\x12@\n\rscript_config\x18\x03 \x01(\x0b\x32).shiftcrypto.bitbox02.CardanoScriptConfig\x12T\n\x0c\x61sset_groups\x18\x04 \x03(\x0b\x32>.shiftcrypto.bitbox02.CardanoSignTransactionRequest.AssetGroup\x1a\xc3\x05\n\x0b\x43\x65rtificate\x12;\n\x12stake_registration\x18\x01 \x01(\x0b\x32\x1d.shiftcrypto.bitbox02.KeypathH\x00\x12=\n\x14stake_deregistration\x18\x02 \x01(\x0b\x32\x1d.shiftcrypto.bitbox02.KeypathH\x00\x12k\n\x10stake_delegation\x18\x03 \x01(\x0b\x32O.shiftcrypto.bitbox02.CardanoSignTransactionRequest.Certificate.StakeDelegationH\x00\x12i\n\x0fvote_delegation\x18\n \x01(\x0b\x32N.shiftcrypto.bitbox02.CardanoSignTransactionRequest.Certificate.VoteDelegationH\x00\x1a\x38\n\x0fStakeDelegation\x12\x0f\n\x07keypath\x18\x01 \x03(\r\x12\x14\n\x0cpool_keyhash\x18\x02 \x01(\x0c\x1a\x9d\x02\n\x0eVoteDelegation\x12\x0f\n\x07keypath\x18\x01 \x03(\r\x12l\n\x04type\x18\x02 \x01(\x0e\x32^.shiftcrypto.bitbox02.CardanoSignTransactionRequest.Certificate.VoteDelegation.CardanoDRepType\x12\x1a\n\rdrep_credhash\x18\x03 \x01(\x0cH\x00\x88\x01\x01\"^\n\x0f\x43\x61rdanoDRepType\x12\x0c\n\x08KEY_HASH\x10\x00\x12\x0f\n\x0bSCRIPT_HASH\x10\x01\x12\x12\n\x0e\x41LWAYS_ABSTAIN\x10\x02\x12\x18\n\x14\x41LWAYS_NO_CONFIDENCE\x10\x03\x42\x10\n\x0e_drep_credhashB\x06\n\x04\x63\x65rt\x1a,\n\nWithdrawal\x12\x0f\n\x07keypath\x18\x01 \x03(\r\x12\r\n\x05value\x18\x02 \x01(\x04\"\xb9\x01\n\x1e\x43\x61rdanoSignTransactionResponse\x12^\n\x11shelley_witnesses\x18\x01 \x03(\x0b\x32\x43.shiftcrypto.bitbox02.CardanoSignTransactionResponse.ShelleyWitness\x1a\x37\n\x0eShelleyWitness\x12\x12\n\npublic_key\x18\x01 \x01(\x0c\x12\x11\n\tsignature\x18\x02 \x01(\x0c\"\xe8\x01\n\x0e\x43\x61rdanoRequest\x12:\n\x05xpubs\x18\x01 \x01(\x0b\x32).shiftcrypto.bitbox02.CardanoXpubsRequestH\x00\x12>\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32+.shiftcrypto.bitbox02.CardanoAddressRequestH\x00\x12O\n\x10sign_transaction\x18\x03 \x01(\x0b\x32\x33.shiftcrypto.bitbox02.CardanoSignTransactionRequestH\x00\x42\t\n\x07request\"\xde\x01\n\x0f\x43\x61rdanoResponse\x12;\n\x05xpubs\x18\x01 \x01(\x0b\x32*.shiftcrypto.bitbox02.CardanoXpubsResponseH\x00\x12\x30\n\x03pub\x18\x02 \x01(\x0b\x32!.shiftcrypto.bitbox02.PubResponseH\x00\x12P\n\x10sign_transaction\x18\x03 \x01(\x0b\x32\x34.shiftcrypto.bitbox02.CardanoSignTransactionResponseH\x00\x42\n\n\x08response*8\n\x0e\x43\x61rdanoNetwork\x12\x12\n\x0e\x43\x61rdanoMainnet\x10\x00\x12\x12\n\x0e\x43\x61rdanoTestnet\x10\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'cardano_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CARDANONETWORK._serialized_start=2852
  _CARDANONETWORK._serialized_end=2908
  _CARDANOXPUBSREQUEST._serialized_start=53
  _CARDANOXPUBSREQUEST._serialized_end=123
  _CARDANOXPUBSRESPONSE._serialized_start=125
  _CARDANOXPUBSRESPONSE._serialized_end=162
  _CARDANOSCRIPTCONFIG._serialized_start=165
  _CARDANOSCRIPTCONFIG._serialized_end=323
  _CARDANOSCRIPTCONFIG_PKHSKH._serialized_start=257
  _CARDANOSCRIPTCONFIG_PKHSKH._serialized_end=313
  _CARDANOADDRESSREQUEST._serialized_start=326
  _CARDANOADDRESSREQUEST._serialized_end=487
  _CARDANOSIGNTRANSACTIONREQUEST._serialized_start=490
  _CARDANOSIGNTRANSACTIONREQUEST._serialized_end=2202
  _CARDANOSIGNTRANSACTIONREQUEST_INPUT._serialized_start=1008
  _CARDANOSIGNTRANSACTIONREQUEST_INPUT._serialized_end=1079
  _CARDANOSIGNTRANSACTIONREQUEST_ASSETGROUP._serialized_start=1082
  _CARDANOSIGNTRANSACTIONREQUEST_ASSETGROUP._serialized_end=1243
  _CARDANOSIGNTRANSACTIONREQUEST_ASSETGROUP_TOKEN._serialized_start=1201
  _CARDANOSIGNTRANSACTIONREQUEST_ASSETGROUP_TOKEN._serialized_end=1243
  _CARDANOSIGNTRANSACTIONREQUEST_OUTPUT._serialized_start=1246
  _CARDANOSIGNTRANSACTIONREQUEST_OUTPUT._serialized_end=1446
  _CARDANOSIGNTRANSACTIONREQUEST_CERTIFICATE._serialized_start=1449
  _CARDANOSIGNTRANSACTIONREQUEST_CERTIFICATE._serialized_end=2156
  _CARDANOSIGNTRANSACTIONREQUEST_CERTIFICATE_STAKEDELEGATION._serialized_start=1804
  _CARDANOSIGNTRANSACTIONREQUEST_CERTIFICATE_STAKEDELEGATION._serialized_end=1860
  _CARDANOSIGNTRANSACTIONREQUEST_CERTIFICATE_VOTEDELEGATION._serialized_start=1863
  _CARDANOSIGNTRANSACTIONREQUEST_CERTIFICATE_VOTEDELEGATION._serialized_end=2148
  _CARDANOSIGNTRANSACTIONREQUEST_CERTIFICATE_VOTEDELEGATION_CARDANODREPTYPE._serialized_start=2036
  _CARDANOSIGNTRANSACTIONREQUEST_CERTIFICATE_VOTEDELEGATION_CARDANODREPTYPE._serialized_end=2130
  _CARDANOSIGNTRANSACTIONREQUEST_WITHDRAWAL._serialized_start=2158
  _CARDANOSIGNTRANSACTIONREQUEST_WITHDRAWAL._serialized_end=2202
  _CARDANOSIGNTRANSACTIONRESPONSE._serialized_start=2205
  _CARDANOSIGNTRANSACTIONRESPONSE._serialized_end=2390
  _CARDANOSIGNTRANSACTIONRESPONSE_SHELLEYWITNESS._serialized_start=2335
  _CARDANOSIGNTRANSACTIONRESPONSE_SHELLEYWITNESS._serialized_end=2390
  _CARDANOREQUEST._serialized_start=2393
  _CARDANOREQUEST._serialized_end=2625
  _CARDANORESPONSE._serialized_start=2628
  _CARDANORESPONSE._serialized_end=2850
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: keystore.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0ekeystore.proto\x12\x14shiftcrypto.bitbox02\x1a\x1bgoogle/protobuf/empty.proto\"/\n\x1c\x45lectrumEncryptionKeyRequest\x12\x0f\n\x07keypath\x18\x01 \x03(\r\",\n\x1d\x45lectrumEncryptionKeyResponse\x12\x0b\n\x03key\x18\x01 \x01(\t\"\x97\x01\n\x0c\x42IP85Request\x12\'\n\x05\x62ip39\x18\x01 \x01(\x0b\x32\x16.google.protobuf.EmptyH\x00\x12\x36\n\x02ln\x18\x02 \x01(\x0b\x32(.shiftcrypto.bitbox02.BIP85Request.AppLnH\x00\x1a\x1f\n\x05\x41ppLn\x12\x16\n\x0e\x61\x63\x63ount_number\x18\x01 \x01(\rB\x05\n\x03\x61pp\"M\n\rBIP85Response\x12\'\n\x05\x62ip39\x18\x01 \x01(\x0b\x32\x16.google.protobuf.EmptyH\x00\x12\x0c\n\x02ln\x18\x02 \x01(\x0cH\x00\x42\x05\n\x03\x61ppb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'keystore_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _ELECTRUMENCRYPTIONKEYREQUEST._serialized_start=69
  _ELECTRUMENCRYPTIONKEYREQUEST._serialized_end=116
  _ELECTRUMENCRYPTIONKEYRESPONSE._serialized_start=118
  _ELECTRUMENCRYPTIONKEYRESPONSE._serialized_end=162
  _BIP85REQUEST._serialized_start=165
  _BIP85REQUEST._serialized_end=316
  _BIP85REQUEST_APPLN._serialized_start=278
  _BIP85REQUEST_APPLN._serialized_end=309
  _BIP85RESPONSE._serialized_start=318
  _BIP85RESPONSE._serialized_end=395
# @@protoc_insertion_point(module_scope)



# Copyright 2019 Shift Cryptosecurity AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common interfaces to be used for communication with BitBox devices.
"""

from typing_extensions import Protocol


class TransportLayer(Protocol):
    """
    Abstraction for the transport layer used for transmitting U2F messages.
    This class encapsulates packets on a given physical link capable of
    transmitting byte strings.
    """

    # pylint: disable=unused-argument
    def write(self, data: bytes, endpoint: int, cid: int) -> None:
        """Sends a frame of data to the specified endpoint"""

    def read(self, endpoint: int, cid: int) -> bytes: ...

    def query(self, data: bytes, endpoint: int, cid: int) -> bytes:
        self.write(data, endpoint, cid)
        return self.read(endpoint, cid)

    def generate_cid(self) -> int: ...

    def close(self) -> None: ...


class PhysicalLayer(Protocol):
    # pylint: disable=unused-argument
    def write(self, data: bytes) -> None: ...

    def read(self, size: int, timeout_ms: int) -> bytes: ...

    def close(self) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2019 Shift Cryptosecurity AG

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class PubResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    PUB_FIELD_NUMBER: builtins.int
    pub: builtins.str
    def __init__(
        self,
        *,
        pub: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["pub", b"pub"]) -> None: ...

global___PubResponse = PubResponse

@typing.final
class RootFingerprintRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___RootFingerprintRequest = RootFingerprintRequest

@typing.final
class RootFingerprintResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FINGERPRINT_FIELD_NUMBER: builtins.int
    fingerprint: builtins.bytes
    def __init__(
        self,
        *,
        fingerprint: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["fingerprint", b"fingerprint"]) -> None: ...

global___RootFingerprintResponse = RootFingerprintResponse

@typing.final
class XPub(google.protobuf.message.Message):
    """See https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki.
    version field dropped as it will set dynamically based on the context (xpub, ypub, etc.).
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DEPTH_FIELD_NUMBER: builtins.int
    PARENT_FINGERPRINT_FIELD_NUMBER: builtins.int
    CHILD_NUM_FIELD_NUMBER: builtins.int
    CHAIN_CODE_FIELD_NUMBER: builtins.int
    PUBLIC_KEY_FIELD_NUMBER: builtins.int
    depth: builtins.bytes
    parent_fingerprint: builtins.bytes
    child_num: builtins.int
    chain_code: builtins.bytes
    public_key: builtins.bytes
    def __init__(
        self,
        *,
        depth: builtins.bytes = ...,
        parent_fingerprint: builtins.bytes = ...,
        child_num: builtins.int = ...,
        chain_code: builtins.bytes = ...,
        public_key: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["chain_code", b"chain_code", "child_num", b"child_num", "depth", b"depth", "parent_fingerprint", b"parent_fingerprint", "public_key", b"public_key"]) -> None: ...

global___XPub = XPub

@typing.final
class Keypath(google.protobuf.message.Message):
    """This message exists for use in oneof or repeated fields, where one can't inline `repeated uint32` due to protobuf rules."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEYPATH_FIELD_NUMBER: builtins.int
    @property
    def keypath(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        keypath: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["keypath", b"keypath"]) -> None: ...

global___Keypath = Keypath

@typing.final
class KeyOriginInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ROOT_FINGERPRINT_FIELD_NUMBER: builtins.int
    KEYPATH_FIELD_NUMBER: builtins.int
    XPUB_FIELD_NUMBER: builtins.int
    root_fingerprint: builtins.bytes
    @property
    def keypath(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    @property
    def xpub(self) -> global___XPub: ...
    def __init__(
        self,
        *,
        root_fingerprint: builtins.bytes = ...,
        keypath: collections.abc.Iterable[builtins.int] | None = ...,
        xpub: global___XPub | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["xpub", b"xpub"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["keypath", b"keypath", "root_fingerprint", b"root_fingerprint", "xpub", b"xpub"]) -> None: ...

global___KeyOriginInfo = KeyOriginInfo



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: bluetooth.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0f\x62luetooth.proto\x12\x14shiftcrypto.bitbox02\"\x1f\n\x1d\x42luetoothToggleEnabledRequest\"6\n\x1b\x42luetoothUpgradeInitRequest\x12\x17\n\x0f\x66irmware_length\x18\x01 \x01(\r\"%\n\x15\x42luetoothChunkRequest\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\x12\n\x10\x42luetoothSuccess\"?\n\x1d\x42luetoothRequestChunkResponse\x12\x0e\n\x06offset\x18\x01 \x01(\r\x12\x0e\n\x06length\x18\x02 \x01(\r\"\xf5\x01\n\x10\x42luetoothRequest\x12I\n\x0cupgrade_init\x18\x01 \x01(\x0b\x32\x31.shiftcrypto.bitbox02.BluetoothUpgradeInitRequestH\x00\x12<\n\x05\x63hunk\x18\x02 \x01(\x0b\x32+.shiftcrypto.bitbox02.BluetoothChunkRequestH\x00\x12M\n\x0etoggle_enabled\x18\x03 \x01(\x0b\x32\x33.shiftcrypto.bitbox02.BluetoothToggleEnabledRequestH\x00\x42\t\n\x07request\"\xa8\x01\n\x11\x42luetoothResponse\x12\x39\n\x07success\x18\x01 \x01(\x0b\x32&.shiftcrypto.bitbox02.BluetoothSuccessH\x00\x12L\n\rrequest_chunk\x18\x02 \x01(\x0b\x32\x33.shiftcrypto.bitbox02.BluetoothRequestChunkResponseH\x00\x42\n\n\x08responseb\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'bluetooth_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _BLUETOOTHTOGGLEENABLEDREQUEST._serialized_start=41
  _BLUETOOTHTOGGLEENABLEDREQUEST._serialized_end=72
  _BLUETOOTHUPGRADEINITREQUEST._serialized_start=74
  _BLUETOOTHUPGRADEINITREQUEST._serialized_end=128
  _BLUETOOTHCHUNKREQUEST._serialized_start=130
  _BLUETOOTHCHUNKREQUEST._serialized_end=167
  _BLUETOOTHSUCCESS._serialized_start=169
  _BLUETOOTHSUCCESS._serialized_end=187
  _BLUETOOTHREQUESTCHUNKRESPONSE._serialized_start=189
  _BLUETOOTHREQUESTCHUNKRESPONSE._serialized_end=252
  _BLUETOOTHREQUEST._serialized_start=255
  _BLUETOOTHREQUEST._serialized_end=500
  _BLUETOOTHRESPONSE._serialized_start=503
  _BLUETOOTHRESPONSE._serialized_end=671
# @@protoc_insertion_point(module_scope)
